# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definitions of various standard energy functions."""

from functools import wraps, partial

from typing import Callable, Tuple, TextIO, Dict, Any, Optional

import jax
import jax.numpy as jnp
from jax import ops
from jax.tree_util import tree_map
from jax import vmap
import haiku as hk
from jax.scipy.special import erfc  # error function
from jax_md import space, smap, partition, nn, quantity, interpolate, util

maybe_downcast = util.maybe_downcast

# Types


f32 = util.f32
f64 = util.f64
Array = util.Array

PyTree = Any
Box = space.Box
DisplacementFn = space.DisplacementFn
DisplacementOrMetricFn = space.DisplacementOrMetricFn

NeighborFn = partition.NeighborFn
NeighborList = partition.NeighborList
NeighborListFormat = partition.NeighborListFormat


# Energy Functions


def simple_spring(dr: Array,
                  length: Array=1,
                  epsilon: Array=1,
                  alpha: Array=2,
                  **unused_kwargs) -> Array:
  """Isotropic spring potential with a given rest length.

  We define `simple_spring` to be a generalized Hookean spring with
  agreement when `alpha = 2`.
  """
  return epsilon / alpha * jnp.abs(dr - length) ** alpha


def simple_spring_bond(displacement_or_metric: DisplacementOrMetricFn,
                       bond: Array,
                       bond_type: Optional[Array]=None,
                       length: Array=1,
                       epsilon: Array=1,
                       alpha: Array=2) -> Callable[[Array], Array]:
  """Convenience wrapper to compute energy of particles bonded by springs."""
  length = maybe_downcast(length)
  epsilon = maybe_downcast(epsilon)
  alpha = maybe_downcast(alpha)
  return smap.bond(
    simple_spring,
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    bond,
    bond_type,
    ignore_unused_parameters=True,
    length=length,
    epsilon=epsilon,
    alpha=alpha)


def soft_sphere(dr: Array,
                sigma: Array=1,
                epsilon: Array=1,
                alpha: Array=2,
                **unused_kwargs) -> Array:
  """.. _soft-sphere:

  Finite ranged repulsive interaction between soft spheres.

  Args:
    dr: An ndarray of shape `[n, m]` of pairwise distances between particles.
    sigma: Particle diameter. Should either be a floating point scalar or an
      ndarray whose shape is `[n, m]`.
    epsilon: Interaction energy scale. Should either be a floating point scalar
      or an ndarray whose shape is `[n, m]`.
    alpha: Exponent specifying interaction stiffness. Should either be a float
      point scalar or an ndarray whose shape is `[n, m]`.
    unused_kwargs: Allows extra data (e.g. time) to be passed to the energy.
  Returns:
    Matrix of energies whose shape is `[n, m]`.
  """

  dr = dr / sigma
  fn = lambda dr: epsilon / alpha * (f32(1.0) - dr) ** alpha

  if isinstance(alpha, int) or issubclass(type(alpha.dtype), jnp.integer):
    return jnp.where(dr < 1.0, fn(dr), f32(0.0))

  return util.safe_mask(dr < 1.0, fn, dr, f32(0.0))


def soft_sphere_pair(displacement_or_metric: DisplacementOrMetricFn,
                     species: Optional[Array]=None,
                     sigma: Array=1.0,
                     epsilon: Array=1.0,
                     alpha: Array=2.0,
                     per_particle: bool=False) -> Callable[[Array], Array]:
  """Convenience wrapper to compute :ref:`soft sphere energy <soft-sphere>` over a system."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  alpha = maybe_downcast(alpha)
  return smap.pair(
      soft_sphere,
      space.canonicalize_displacement_or_metric(displacement_or_metric),
      ignore_unused_parameters=True,
      species=species,
      sigma=sigma,
      epsilon=epsilon,
      alpha=alpha,
      reduce_axis=(1,) if per_particle else None)


def soft_sphere_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: Box,
    species: Optional[Array]=None,
    sigma: Array=1.0,
    epsilon: Array=1.0,
    alpha: Array=2.0,
    dr_threshold: float=0.2,
    per_particle: bool=False,
    fractional_coordinates: bool=False,
    format: NeighborListFormat=partition.OrderedSparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`soft spheres <soft-sphere>` using a neighbor list."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  alpha = maybe_downcast(alpha)
  list_cutoff = jnp.max(sigma)
  dr_threshold = maybe_downcast(dr_threshold)

  neighbor_fn = partition.neighbor_list(
    displacement_or_metric,
    box_size,
    list_cutoff,
    dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)
  energy_fn = smap.pair_neighbor_list(
    soft_sphere,
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    ignore_unused_parameters=True,
    species=species,
    sigma=sigma,
    epsilon=epsilon,
    alpha=alpha,
    reduce_axis=(1,) if per_particle else None)

  return neighbor_fn, energy_fn


def lennard_jones(dr: Array,
                  sigma: Array=1,
                  epsilon: Array=1,
                  **unused_kwargs) -> Array:
  """.. _lj-pot:
  
  Lennard-Jones interaction between particles with a minimum at `sigma`.

  Args:
    dr: An ndarray of shape `[n, m]` of pairwise distances between particles.
    sigma: Distance between particles where the energy has a minimum. Should
      either be a floating point scalar or an ndarray whose shape is `[n, m]`.
    epsilon: Interaction energy scale. Should either be a floating point scalar
      or an ndarray whose shape is `[n, m]`.
    unused_kwargs: Allows extra data (e.g. time) to be passed to the energy.
  Returns:
    Matrix of energies of shape `[n, m]`.
  """
  idr = (sigma / dr)
  idr = idr * idr
  idr6 = idr * idr * idr
  idr12 = idr6 * idr6
  # TODO(schsam): This seems potentially dangerous. We should do ErrorChecking
  # here.
  return jnp.nan_to_num(f32(4) * epsilon * (idr12 - idr6))


def lennard_jones_pair(displacement_or_metric: DisplacementOrMetricFn,
                       species: Optional[Array]=None,
                       sigma: Array=1.0,
                       epsilon: Array=1.0,
                       r_onset: Array=2.0,
                       r_cutoff: Array=2.5,
                       per_particle: bool=False) -> Callable[[Array], Array]:
  """Convenience wrapper to compute :ref:`Lennard-Jones energy <lj-pot>` over a system."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  r_onset = maybe_downcast(r_onset) * jnp.max(sigma)
  r_cutoff = maybe_downcast(r_cutoff) * jnp.max(sigma)
  return smap.pair(
    multiplicative_isotropic_cutoff(lennard_jones, r_onset, r_cutoff),
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    ignore_unused_parameters=True,
    species=species,
    sigma=sigma,
    epsilon=epsilon,
    reduce_axis=(1,) if per_particle else None)


def lennard_jones_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: Box,
    species: Optional[Array]=None,
    sigma: Array=1.0,
    epsilon: Array=1.0,
    alpha: Array=2.0,
    r_onset: float=2.0,
    r_cutoff: float=2.5,
    dr_threshold: float=0.5,
    per_particle: bool=False,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.OrderedSparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`Lennard-Jones <lj-pot>` using a neighbor list."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  r_onset = maybe_downcast(r_onset) * jnp.max(sigma)
  r_cutoff = maybe_downcast(r_cutoff) * jnp.max(sigma)
  dr_threshold = maybe_downcast(dr_threshold)

  neighbor_fn = partition.neighbor_list(
    displacement_or_metric,
    box_size,
    r_cutoff,
    dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)
  energy_fn = smap.pair_neighbor_list(
    multiplicative_isotropic_cutoff(lennard_jones, r_onset, r_cutoff),
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    ignore_unused_parameters=True,
    species=species,
    sigma=sigma,
    epsilon=epsilon,
    reduce_axis=(1,) if per_particle else None)

  return neighbor_fn, energy_fn


def morse(dr: Array,
          sigma: Array=1.0,
          epsilon: Array=5.0,
          alpha: Array=5.0,
          **unused_kwargs) -> Array:
  """.. _morse-pot:
  
  Morse interaction between particles with a minimum at `sigma`.

  Args:
    dr: An ndarray of shape `[n, m]` of pairwise distances between particles.
    sigma: Distance between particles where the energy has a minimum. Should
      either be a floating point scalar or an ndarray whose shape is `[n, m]`.
    epsilon: Interaction energy scale. Should either be a floating point scalar
      or an ndarray whose shape is `[n, m]`.
    alpha: Range parameter. Should either be a floating point scalar or an
      ndarray whose shape is `[n, m]`.
    unused_kwargs: Allows extra data (e.g. time) to be passed to the energy.
  Returns:
    Matrix of energies of shape `[n, m]`.
  """
  U = epsilon * (f32(1) - jnp.exp(-alpha * (dr - sigma)))**f32(2) - epsilon
  # TODO(cpgoodri): ErrorChecking following lennard_jones
  return jnp.nan_to_num(jnp.array(U, dtype=dr.dtype))


def morse_pair(displacement_or_metric: DisplacementOrMetricFn,
               species: Optional[Array]=None,
               sigma: Array=1.0,
               epsilon: Array=5.0,
               alpha: Array=5.0,
               r_onset: float=2.0,
               r_cutoff: float=2.5,
               per_particle: bool=False) -> Callable[[Array], Array]:
  """Convenience wrapper to compute :ref:`Morse energy <morse-pot>` over a system."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  alpha = maybe_downcast(alpha)
  return smap.pair(
    multiplicative_isotropic_cutoff(morse, r_onset, r_cutoff),
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    ignore_unused_parameters=True,
    species=species,
    sigma=sigma,
    epsilon=epsilon,
    alpha=alpha,
    reduce_axis=(1,) if per_particle else None)


def morse_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: Box,
    species: Optional[Array]=None,
    sigma: Array=1.0,
    epsilon: Array=5.0,
    alpha: Array=5.0,
    r_onset: float=2.0,
    r_cutoff: float=2.5,
    dr_threshold: float=0.5,
    per_particle: bool=False,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.OrderedSparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`Morse <morse-pot>` using a neighbor list."""
  sigma = maybe_downcast(sigma)
  epsilon = maybe_downcast(epsilon)
  alpha = maybe_downcast(alpha)
  r_onset = maybe_downcast(r_onset)
  r_cutoff = maybe_downcast(r_cutoff)
  dr_threshold = maybe_downcast(dr_threshold)

  neighbor_fn = partition.neighbor_list(
    displacement_or_metric,
    box_size,
    r_cutoff,
    dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)
  energy_fn = smap.pair_neighbor_list(
    multiplicative_isotropic_cutoff(morse, r_onset, r_cutoff),
    space.canonicalize_displacement_or_metric(displacement_or_metric),
    ignore_unused_parameters=True,
    species=species,
    sigma=sigma,
    epsilon=epsilon,
    alpha=alpha,
    reduce_axis=(1,) if per_particle else None)

  return neighbor_fn, energy_fn


def gupta_potential(displacement, p, q, r_0n, U_n, A, cutoff):
  """.. _gupta-pot:

  Gupta potential with default parameters for Au_55 cluster. Gupta
  potential was introduced by R. P. Gupta [#gupta]_. This potential uses parameters
  that were fit for bulk gold by Jellinek [#jellinek]_. This particular implementation
  of the Gupta potential was introduced by Garzon and Posada-Amarillas [#garzon]_.

  Args:
    displacement: Function to compute displacement between two positions.
    p: Gupta potential parameter of the repulsive term that was fitted for bulk gold.
    q: Gupta potential parameter of the attractive term that was fitted for bulk gold.
    r_0n: 
      Parameter that determines the length scale of the potential. This
      value was particularly fit for gold clusters of size 55 atoms.
    U_n: 
      Parameter that determines the energy scale, fit particularly for
      gold clusters of size 55 atoms.
    A: Parameter that was obtained using the cohesive energy of the fcc gold metal.
    cutoff: 
      Pairwise interactions that are farther than the cutoff distance will be ignored.

  Returns:
    A function that takes in positions of gold atoms (shape `[n, 3]` where `n` is
    the number of atoms) and returns the total energy of the system in units
    of eV.

  .. rubric:: References
  .. [#gupta] R.P. Gupta, Phys. Rev. B 23, 6265 (1981)
  .. [#jellinek] J. Jellinek, in Metal-Ligand Interactions, edited by N. Russo and
    D. R. Salahub (Kluwer Academic, Dordrecht, 1996), p. 325.
  .. [#garzon] I.L. Garzon, A. Posada-Amarillas, Phys. Rev. B 54, 16 (1996)
  """
  def _gupta_term1(r, p, r_0n, cutoff):
    """Repulsive term in Gupta potential."""
    within_cutoff = (r > 0) & (r < cutoff)
    term1 = jnp.exp(-1.0 * p * (r / r_0n - 1))
    return jnp.where(within_cutoff, term1, 0.0)

  def _gupta_term2(r, q, r_0n, cutoff):
    """Attractive term in Gupta potential."""
    within_cutoff = (r > 0) & (r < cutoff)
    term2 = jnp.exp(-2.0 * q * (r / r_0n - 1))
    return jnp.where(within_cutoff, term2, 0.0)

  def compute_fn(R):
    dR = space.map_product(displacement)(R, R)
    dr = space.distance(dR)
    first_term = A * jnp.sum(_gupta_term1(dr, p, r_0n, cutoff), axis=1)
    # Safe sqrt used in order to ensure that force calculations are not nan
    # when the particles are too widely separated at initialization
    # (corresponding to the case where the attractive term is 0.).
    attractive_term = jnp.sum(_gupta_term2(dr, q, r_0n, cutoff), axis=1)
    second_term = util.safe_mask(attractive_term > 0,
                                 jnp.sqrt, attractive_term)
    return U_n / 2.0 * jnp.sum(first_term - second_term)

  return compute_fn


GUPTA_GOLD55_DICT = {
    'p' : 10.15,
    'q' : 4.13,
    'r_0n' : 2.96,
    'U_n' : 3.454,
    'A' : 0.118428,
}


def gupta_gold55(displacement,
                 cutoff=8.0):
  gupta_gold_fn = gupta_potential(displacement,
                                  cutoff=cutoff,
                                  **GUPTA_GOLD55_DICT)
  def energy_fn(R, **unused_kwargs):
    return gupta_gold_fn(R)
  return energy_fn


def multiplicative_isotropic_cutoff(fn: Callable[..., Array],
                                    r_onset: float,
                                    r_cutoff: float) -> Callable[..., Array]:
  """Takes an isotropic function and constructs a truncated function.

  Given a function `f:R -> R`, we construct a new function `f':R -> R` such that
  `f'(r) = f(r)` for `r < r_onset`, `f'(r) = 0` for `r > r_cutoff`, and `f(r)` is :math:`C^1`
  everywhere. To do this, we follow the approach outlined in HOOMD Blue  [#hoomd]_
  (thanks to Carl Goodrich for the pointer). We construct a function `S(r)` such
  that `S(r) = 1` for `r < r_onset`, `S(r) = 0` for `r > r_cutoff`, and `S(r)` is :math:`C^1`.
  Then `f'(r) = S(r)f(r)`.

  Args:
    fn: A function that takes an ndarray of distances of shape `[n, m]` as well
      as varargs.
    r_onset: A float specifying the distance marking the onset of deformation.
    r_cutoff: A float specifying the cutoff distance.

  Returns:
    A new function with the same signature as fn, with the properties outlined
    above.

  .. rubric:: References
  .. [#hoomd] HOOMD Blue documentation. Accessed on 05/31/2019.
      https://hoomd-blue.readthedocs.io/en/stable/module-md-pair.html#hoomd.md.pair.pair
  """

  r_c = r_cutoff ** f32(2)
  r_o = r_onset ** f32(2)

  def smooth_fn(dr):
    r = dr ** f32(2)

    inner = jnp.where(dr < r_cutoff,
                     (r_c - r)**2 * (r_c + 2 * r - 3 * r_o) / (r_c - r_o)**3,
                     0)

    return jnp.where(dr < r_onset, 1, inner)

  @wraps(fn)
  def cutoff_fn(dr, *args, **kwargs):
    return smooth_fn(dr) * fn(dr, *args, **kwargs)

  return cutoff_fn


def dsf_coulomb(r: Array,
                Q_sq: Array,
                alpha: Array=0.25,
                cutoff: float=8.0) -> Array:
  """Damped-shifted-force approximation of the coulombic interaction."""
  qqr2e = 332.06371  # Coulombic conversion factor: 1/(4*pi*epo).

  cutoffsq = cutoff * cutoff
  erfcc = erfc(alpha * cutoff)
  erfcd = jnp.exp(-alpha * alpha * cutoffsq)
  f_shift = -(erfcc / cutoffsq + 2 / jnp.sqrt(jnp.pi) * alpha * erfcd / cutoff)
  e_shift = erfcc / cutoff - f_shift * cutoff

  e = qqr2e * Q_sq / r * (erfc(alpha * r) - r * e_shift - r**2 * f_shift)
  return jnp.where(r < cutoff, e, 0.0)


def bks(dr: Array,
        Q_sq: Array,
        exp_coeff: Array,
        exp_decay: Array,
        attractive_coeff: Array,
        repulsive_coeff: Array,
        coulomb_alpha: Array,
        cutoff: float,
        **unused_kwargs) -> Array:
  """.. _bks-pot:
  
  Beest-Kramer-van Santen (BKS) potential [#bks]_ which is commonly used to
  model silicas. This function computes the interaction between two
  given atoms within the Buckingham form [#carre]_ , following the implementation
  from Liu et al. [#liu]_ .

  Args:
    dr: An ndarray of shape `[n, m]` of pairwise distances between particles.
    Q_sq: An ndarray of shape `[n, m]` of pairwise product of partial charges.
    exp_coeff: An ndarray of shape `[n, m]` that sets the scale of the
      exponential decay of the short-range interaction.
    attractive_coeff: An ndarray of shape `[n, m]` for the coefficient of the
      attractive 6th order term.
    repulsive_coeff: An ndarray of shape `[n, m]` for the coefficient of the
      repulsive 24th order term, to prevent the unphysical fusion of atoms.
    coulomb_alpha: Damping parameter for the approximation of the long-range
      coulombic interactions (a scalar).
    cutoff: Cutoff distance for considering pairwise interactions.
    unused_kwargs: Allows extra data (e.g. time) to be passed to the energy.

  Returns:
    Matrix of energies of shape `[n, m]`.

  .. rubric:: References
  .. [#bks] Van Beest, B. W. H., Gert Jan Kramer, and R. A. Van Santen. "Force fields
    for silicas and aluminophosphates based on ab initio calculations." Physical
    Review Letters 64.16 (1990): 1955.
  .. [#carre] Carré, Antoine, et al. "Developing empirical potentials from ab initio
    simulations: The case of amorphous silica." Computational Materials Science
    124 (2016): 323-334.
  .. [#liu] Liu, Han, et al. "Machine learning Forcefield for silicate glasses."
    arXiv preprint arXiv:1902.03486 (2019).
  """
  energy = (dsf_coulomb(dr, Q_sq, coulomb_alpha, cutoff) + \
            exp_coeff * jnp.exp(-dr / exp_decay) + \
            attractive_coeff / dr ** 6 + repulsive_coeff / dr ** 24)
  return  jnp.where(dr < cutoff, energy, 0.0)


def bks_pair(displacement_or_metric: DisplacementOrMetricFn,
             species: Array,
             Q_sq: Array,
             exp_coeff: Array,
             exp_decay: Array,
             attractive_coeff: Array,
             repulsive_coeff: Array,
             coulomb_alpha: Array,
             cutoff: float) -> Callable[[Array], Array]:
  """Convenience wrapper to compute :ref:`BKS energy <bks-pot>` over a system."""
  Q_sq = maybe_downcast(Q_sq)
  exp_coeff = maybe_downcast(exp_coeff)
  exp_decay = maybe_downcast(exp_decay)
  attractive_coeff = maybe_downcast(attractive_coeff)
  repulsive_coeff = maybe_downcast(repulsive_coeff)

  return smap.pair(bks, displacement_or_metric,
                   species=species,
                   ignore_unused_parameters=True,
                   Q_sq=Q_sq,
                   exp_coeff=exp_coeff,
                   exp_decay=exp_decay,
                   attractive_coeff=attractive_coeff,
                   repulsive_coeff=repulsive_coeff,
                   coulomb_alpha=coulomb_alpha,
                   cutoff=cutoff)


def bks_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: Box,
    species: Array,
    Q_sq: Array,
    exp_coeff: Array,
    exp_decay: Array,
    attractive_coeff: Array,
    repulsive_coeff: Array,
    coulomb_alpha: Array,
    cutoff: float,
    dr_threshold: float=0.8,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.OrderedSparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`BKS energy <bks-pot>` using a neighbor list."""
  Q_sq = maybe_downcast(Q_sq)
  exp_coeff = maybe_downcast(exp_coeff)
  exp_decay = maybe_downcast(exp_decay)
  attractive_coeff = maybe_downcast(attractive_coeff)
  repulsive_coeff = maybe_downcast(repulsive_coeff)
  dr_threshold = maybe_downcast(dr_threshold)

  neighbor_fn = partition.neighbor_list(
    displacement_or_metric,
    box_size,
    cutoff,
    dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)

  energy_fn = smap.pair_neighbor_list(
      bks,
      space.canonicalize_displacement_or_metric(displacement_or_metric),
      species=species,
      ignore_unused_parameters=True,
      Q_sq=Q_sq,
      exp_coeff=exp_coeff,
      exp_decay=exp_decay,
      attractive_coeff=attractive_coeff,
      repulsive_coeff=repulsive_coeff,
      coulomb_alpha=coulomb_alpha,
      cutoff=cutoff)

  return neighbor_fn, energy_fn

# BKS Potential Parameters.
# Coefficients given in kcal/mol.

CHARGE_OXYGEN = -0.977476019
CHARGE_SILICON = 1.954952037

BKS_SILICA_DICT = {
    'Q_sq' : [[CHARGE_SILICON**2, CHARGE_SILICON*CHARGE_OXYGEN],
              [CHARGE_SILICON*CHARGE_OXYGEN, CHARGE_OXYGEN**2]],
    'exp_coeff' : [[0, 471671.1243 ],
                   [471671.1243, 23138.64826]],
    'exp_decay' : [[1, 0.19173537],
                   [0.19173537, 0.356855265]],
    'attractive_coeff' : [[0, -2156.074422],
                          [-2156.074422, -1879.223108]],
    'repulsive_coeff' : [[78940848.06, 668.7557239],
                         [668.7557239, 2605.841269]],
    'coulomb_alpha' : 0.25,
}


def _bks_silica_self(Q_sq: Array, alpha: Array, cutoff: float) -> Array:
  """Function for computing the self-energy contributions to BKS."""
  cutoffsq = cutoff * cutoff
  erfcc = erfc(alpha * cutoff)
  erfcd = jnp.exp(-alpha * alpha * cutoffsq)
  f_shift = -(erfcc / cutoffsq + 2.0 / jnp.sqrt(jnp.pi) * alpha * erfcd / cutoff)
  e_shift = erfcc / cutoff - f_shift * cutoff
  qqr2e = 332.06371  # kcal/mol coulombic conversion factor: 1/(4*pi*epo)
  return -(e_shift / 2.0 + alpha / jnp.sqrt(jnp.pi)) * Q_sq * qqr2e


def bks_silica_pair(displacement_or_metric: DisplacementOrMetricFn,
                    species: Array,
                    cutoff: float=8.0):
  """Convenience wrapper to compute :ref:`BKS energy <bks-pot>` for SiO2."""
  bks_pair_fn = bks_pair(displacement_or_metric,
                         species,
                         cutoff=cutoff,
                         **BKS_SILICA_DICT)
  N_0 = jnp.sum(species==0)
  N_1 = jnp.sum(species==1)

  e_self = partial(_bks_silica_self, alpha=0.25, cutoff=cutoff)

  def energy_fn(R, **kwargs):
    return (bks_pair_fn(R, **kwargs) +
            N_0 * e_self(CHARGE_SILICON**2) +
            N_1 * e_self(CHARGE_OXYGEN**2))


  return energy_fn


def bks_silica_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: Box,
    species: Array,
    cutoff: float = 8.0,
    dr_threshold: float = 1.0,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.OrderedSparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`BKS energy <bks-pot>` using neighbor lists."""
  kwargs = {**BKS_SILICA_DICT, **neighbor_kwargs}
  neighbor_fn, bks_pair_fn = bks_neighbor_list(
    displacement_or_metric,
    box_size,
    species,
    cutoff=cutoff,
    dr_threshold=dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **kwargs)
  N_0 = jnp.sum(species==0)
  N_1 = jnp.sum(species==1)

  e_self = partial(_bks_silica_self, alpha=0.25, cutoff=cutoff)

  def energy_fn(R, neighbor, **kwargs):
    return (bks_pair_fn(R, neighbor, **kwargs) +
            N_0 * e_self(CHARGE_SILICON ** 2) +
            N_1 * e_self(CHARGE_OXYGEN ** 2))

  return neighbor_fn, energy_fn


# Stillinger-Weber Potential
def _sw_angle_interaction(gamma: float, sigma: float, cutoff: float,
                          dR12: Array, dR13: Array) -> Array:
  """The angular interaction for the Stillinger-Weber potential.
  This function is defined only for interaction with a pair of
  neighbors. We then vmap this function three times below to make
  it work on the whole system of atoms.
  Args:
    gamma: A scalar used to fit the angle interaction.
    sigma: A scalar that sets the distance scale between neighbors.
    cutoff: The cutoff beyond which the interactions are not
      considered. The default value should not be changed for the
      default SW potential.
    dR12: A d-dimensional vector that specifies the displacement
      of the first neighbor. This potential is usually used in three
      dimensions.
    dR13: A d-dimensional vector that specifies the displacement
      of the second neighbor.

  Returns:
    Angular interaction energy for one pair of neighbors.
  """
  a = cutoff / sigma
  dr12 = space.distance(dR12)
  dr13 = space.distance(dR13)
  dr12 = jnp.where(dr12 < cutoff, dr12, 0)
  dr13 = jnp.where(dr13 < cutoff, dr13, 0)
  term1 = jnp.exp(gamma / (dr12 / sigma - a) + gamma / (dr13 / sigma - a))
  cos_angle = quantity.cosine_angle_between_two_vectors(dR12, dR13)
  term2 = (cos_angle + 1./3)**2
  within_cutoff = (dr12>0) & (dr13>0) & (jnp.linalg.norm(dR12-dR13)>1e-5)
  return jnp.where(within_cutoff, term1 * term2, 0)
sw_three_body_term = vmap(vmap(vmap(
    _sw_angle_interaction, (0, None)), (None, 0)), 0)


def _sw_radial_interaction(sigma: float, B: float, cutoff: float, r: Array
                           ) -> Array:
  """The two body term of the Stillinger-Weber potential."""
  a = cutoff / sigma
  p = 4
  term1 = (B * (r / sigma)**(-p) - 1.0)
  within_cutoff = (r > 0) & (r < cutoff)
  r = jnp.where(within_cutoff, r, 0)
  term2 = jnp.exp(1 / (r / sigma - a))
  return jnp.where(within_cutoff, term1 * term2, 0.0)


def stillinger_weber(displacement: DisplacementFn,
                     sigma: float = 2.0951,
                     A: float = 7.049556277,
                     B: float = 0.6022245584,
                     lam: float = 21.0,
                     gamma: float = 1.2,
                     epsilon: float = 2.16826,
                     three_body_strength: float =1.0,
                     cutoff: float = 3.77118) -> Callable[[Array], Array]:
  """.. _sw-pot:

  Computes the Stillinger-Weber potential.

  The Stillinger-Weber (SW) potential [#stillinger]_ which is commonly used to model
  silicon and similar systems. This function uses the default SW parameters
  from the original paper. The SW potential was originally proposed to
  model diamond in the diamond crystal phase and the liquid phase, and is
  known to give unphysical amorphous configurations [#holender]_ [#barkema]_ . 
  For this reason, we provide a `three_body_strength` parameter. Changing this number to `1.5`
  or `2.0` has been know to produce more physical amorphous phase, preventing
  most atoms from having more than four nearest neighbors. Note that this
  function currently assumes nearest-image-convention.

  Args:
    displacement: The displacement function for the space.
    sigma: A scalar that sets the distance scale between neighbors.
    A: A scalar that determines the scale of two-body term.
    B: A scalar that determines the scale of the :math:`1 / r^p` term.
    lam: A scalar that determines the scale of the three-body term.
    epsilon: A scalar that sets the total energy scale.
    gamma: A scalar used to fit the angle interaction.
    three_body_strength: 
      A scalar that determines the relative strength
      of the angular interaction. Default value is `1.0`, which works well
      for the diamond crystal and liquid phases. `1.5` and `2.0` have been used
      to model amorphous silicon.
  Returns:
    A function that computes the total energy.

  .. rubric:: References
  .. [#stillinger] Stillinger, Frank H., and Thomas A. Weber. "Computer simulation of
    local order in condensed phases of silicon." Physical review B 31.8 (1985): 5262.
  .. [#holender] Holender, J. M., and G. J. Morgan. "Generation of a large structure
    (105 atoms) of amorphous Si using molecular dynamics." Journal of
    Physics: Condensed Matter 3.38 (1991): 7241.
  .. [#barkema] Barkema, G. T., and Normand Mousseau. "Event-based relaxation of
    continuous disordered systems." Physical review letters 77.21 (1996): 4358.
  """
  two_body_fn = partial(_sw_radial_interaction, sigma, B, cutoff)
  three_body_fn = partial(_sw_angle_interaction, gamma, sigma, cutoff)
  three_body_fn = vmap(vmap(vmap(three_body_fn, (0, None)), (None, 0)))

  def compute_fn(R, **kwargs):
    d = partial(displacement, **kwargs)
    dR = space.map_product(d)(R, R)
    dr = space.distance(dR)
    first_term = util.high_precision_sum(two_body_fn(dr)) / 2.0 * A
    second_term = lam *  util.high_precision_sum(three_body_fn(dR, dR)) / 2.0
    return epsilon * (first_term + three_body_strength * second_term)
  return compute_fn


def stillinger_weber_neighbor_list(
    displacement: DisplacementFn,
    box_size: float,
    sigma: float = 2.0951,
    A: float = 7.049556277,
    B: float = 0.6022245584,
    lam: float = 21.0,
    gamma: float = 1.2,
    epsilon: float = 2.16826,
    three_body_strength:float = 1.0,
    cutoff: float = 3.77118,
    dr_threshold: float = 0.5,
    fractional_coordinates: bool=False,
    format: NeighborListFormat=partition.Dense,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`Stillinger-Weber <sw-pot>` 
  using a neighbor list.
  """
  two_body_fn = partial(_sw_radial_interaction, sigma, B, cutoff)
  three_body_fn = partial(_sw_angle_interaction, gamma, sigma, cutoff)

  neighbor_fn = partition.neighbor_list(displacement,
                                        box_size,
                                        cutoff,
                                        dr_threshold,
                                        format=format,
                                        **neighbor_kwargs)

  def compute_fn(R, neighbor, **kwargs):
    d = partial(displacement, **kwargs)
    mask = partition.neighbor_list_mask(neighbor)

    if neighbor.format is partition.Dense:
      _three_body_fn = vmap(vmap(vmap(three_body_fn, (0, None)), (None, 0)))
      dR = space.map_neighbor(d)(R, R[neighbor.idx])
      dr = space.distance(dR)
      first_term = util.high_precision_sum(two_body_fn(dr) * mask) / 2.0 * A
      mask_ijk = mask[:, None, :] * mask[:, :, None]
      second_term = lam *  util.high_precision_sum(
        _three_body_fn(dR, dR) * mask_ijk) / 2.0
    else:
      raise NotImplementedError('Stillinger-Weber potential only implemented '
                                'with Dense neighbor lists.')

    return epsilon * (first_term + three_body_strength * second_term)
  return neighbor_fn, compute_fn



# Embedded Atom Method


def load_lammps_eam_parameters(file: TextIO) -> Tuple[Callable[[Array], Array],
                                                      Callable[[Array], Array],
                                                      Callable[[Array], Array],
                                                      float]:
  """Reads EAM parameters from a LAMMPS file and returns relevant spline fits.

  This function reads single-element EAM potential fit parameters from a file
  in DYNAMO funcl format. In summary, the file contains:

  * Line 1-3: Comments
  * Line 4: Number of elements and the element type
  * Line 5: The number of charge values that the embedding energy is evaluated
    on (`num_drho`), interval between the charge values (`drho`), the number of
    distances the pairwise energy and the charge density is evaluated on (`num_dr`),
    the interval between these distances (`dr`), and the cutoff distance (`cutoff`).

  The lines that come after are the embedding function evaluated on `num_drho`
  charge values, charge function evaluated at `num_dr` distance values, and
  pairwise energy evaluated at `num_dr` distance values. Note that the pairwise
  energy is multiplied by distance (in units of eV x Angstroms). 
  
  For more details of the DYNAMO file format, see:
  https://sites.google.com/a/ncsu.edu/cjobrien/tutorials-and-guides/eam
  
  Args:
    f: File handle for the EAM parameters text file.

  Returns:
    A tuple containing three functions and a cutoff distance.
    
    charge_fn: 
      A function that takes an ndarray of shape `[n, m]` of distances
      between particles and returns a matrix of charge contributions.
    embedding_fn: 
      Function that takes an ndarray of shape `[n]` of charges and
      returns an ndarray of shape `[n]` of the energy cost of embedding an atom
      into the charge.
    pairwise_fn: 
      A function that takes an ndarray of shape `[n, m]` of distances
      and returns an ndarray of shape `[n, m]` of pairwise energies.
    cutoff: 
      Cutoff distance for the `embedding_fn` and `pairwise_fn`.
  """
  raw_text = file.read().split('\n')
  if 'setfl' not in raw_text[0]:
    raise ValueError('File format is incorrect, expected LAMMPS setfl format.')
  temp_params = raw_text[4].split()
  num_drho, num_dr = int(temp_params[0]), int(temp_params[2])
  drho, dr, cutoff = float(temp_params[1]), float(temp_params[3]), float(
      temp_params[4])
  data = maybe_downcast([float(i) for i in raw_text[6:-1]])
  embedding_fn = interpolate.spline(data[:num_drho], drho)
  charge_fn = interpolate.spline(data[num_drho:num_drho + num_dr], dr)
  # LAMMPS EAM parameters file lists pairwise energies after multiplying by
  # distance, in units of eV*Angstrom. We divide the energy by distance below,
  distances = jnp.arange(num_dr) * dr
  # Prevent dividing by zero at zero distance, which will not
  # affect the calculation
  distances = jnp.where(distances == 0, f32(0.001), distances)
  pairwise_fn = interpolate.spline(
      data[num_dr + num_drho:num_drho + 2 * num_dr] / distances,
      dr)
  return charge_fn, embedding_fn, pairwise_fn, cutoff


def eam(displacement: DisplacementFn,
        charge_fn: Callable[[Array], Array],
        embedding_fn: Callable[[Array], Array],
        pairwise_fn: Callable[[Array], Array],
        axis: Optional[Tuple[int, ...]]=None) -> Callable[[Array], Array]:
  """.. _eam-pot:
  
  Interatomic potential as approximated by embedded atom model (EAM).

  This code implements the EAM approximation to interactions between metallic
  atoms. In EAM, the potential energy of an atom is given by two terms: a
  pairwise energy and an embedding energy due to the interaction between the
  atom and background charge density. The EAM potential for a single atomic
  species is often determined by three functions:

  1) Charge density contribution of an atom as a function of distance.
  2) Energy of embedding an atom in the background charge density.
  3) Pairwise energy.

  These three functions are usually provided as spline fits, and we follow the
  implementation and spline fits given by Mishin et al. [#mishin]_
  Note that in current implementation, the three functions listed above 
  can also be expressed by a any function with the correct signature, 
  including neural networks.

  Args:
    displacement: A function that produces an ndarray of shape `[n, m,
      spatial_dimension]` of particle displacements from particle positions
      specified as an ndarray of shape `[n, spatial_dimension]` and `[m,
      spatial_dimension]` respectively.
    box_size: The size of the simulation box.
    charge_fn: A function that takes an ndarray of shape `[n, m]` of distances
      between particles and returns a matrix of charge contributions.
    embedding_fn: Function that takes an ndarray of shape `[n]` of charges and
      returns an ndarray of shape `[n]` of the energy cost of embedding an atom
      into the charge.
    pairwise_fn: A function that takes an ndarray of shape `[n, m]` of distances
      and returns an ndarray of shape `[n, m]` of pairwise energies.
    cutoff: A float specifying the maximum interaction distance.
    dr_threshold: A float specifying the halo in the neighbor list.
    axis: Specifies which axis the total energy should be summed over.
    fractional_coordinates: A boolean specifying whether or not the coordinates
      will be in the unit cube.
    format: The format of the neighbor list.

  Returns:
    A tuple containing a function to build the neighbor list and function that
    computes the EAM energy of a set of atoms with positions given by an
    `[n, spatial_dimension]` ndarray.
  
  .. rubric:: References
  .. [#mishin] Y. Mishin, D. Farkas, M.J. Mehl, DA Papaconstantopoulos, "Interatomic
    potentials for monoatomic metals from experimental data and ab initio
    calculations." Physical Review B, 59 (1999)
  """
  metric = space.map_product(space.metric(displacement))

  def energy(R, **kwargs):
    dr = metric(R, R, **kwargs)
    total_charge = util.high_precision_sum(charge_fn(dr), axis=1)
    embedding_energy = embedding_fn(total_charge)
    pairwise_energy = util.high_precision_sum(smap._diagonal_mask(
        pairwise_fn(dr)), axis=1) / f32(2.0)
    return util.high_precision_sum(
        embedding_energy + pairwise_energy, axis=axis)

  return energy


def eam_from_lammps_parameters(displacement: DisplacementFn,
                               f: TextIO) -> Callable[[Array], Array]:
  """Convenience wrapper to compute :ref:`EAM energy <eam-pot>` with LAMMPS parameters."""
  return eam(displacement, *load_lammps_eam_parameters(f)[:-1])


def eam_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    box_size: float,
    charge_fn: Callable[[Array], Array],
    embedding_fn: Callable[[Array], Array],
    pairwise_fn: Callable[[Array], Array],
    cutoff: float,
    dr_threshold: float = 0.5,
    axis: Optional[Tuple[int, ...]] = None,
    fractional_coordinates: bool = True,
    format: partition.NeighborListFormat = partition.Sparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`EAM <eam-pot>` using a neighbor list."""
  metric = space.canonicalize_displacement_or_metric(displacement_or_metric)

  neighbor_fn = partition.neighbor_list(displacement_or_metric,
                                        box_size,
                                        cutoff,
                                        dr_threshold,
                                        mask_self=False,
                                        format=format,
                                        **neighbor_kwargs)

  def energy_fn(R, neighbor, **kwargs):
    mask = partition.neighbor_list_mask(neighbor)
    self_mask = partition.neighbor_list_mask(neighbor, mask_self=True)

    if neighbor.format is partition.Dense:
      dr = space.map_neighbor(metric)(R, R[neighbor.idx], **kwargs)
      total_charge = util.high_precision_sum(charge_fn(dr) * mask, axis=1)
      embedding_energy = embedding_fn(total_charge)
      pairwise_energy = util.high_precision_sum(pairwise_fn(dr) * self_mask,
                                                axis=1)
    elif neighbor.format is partition.Sparse:
      N = len(R)
      dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]],
                                  **kwargs)
      total_charge = ops.segment_sum(charge_fn(dr) * mask, neighbor.idx[0], N)
      embedding_energy = embedding_fn(total_charge)
      pairwise_energy = ops.segment_sum(pairwise_fn(dr) * self_mask,
                                        neighbor.idx[0], N)
    else:
      raise NotImplementedError('EAM potential not implemented for '
                                'OrderedSparse neighbor lists.')

    return util.high_precision_sum(
      embedding_energy + pairwise_energy / 2.0, axis=axis)

  return neighbor_fn, energy_fn


def eam_from_lammps_parameters_neighbor_list(
    displacement: DisplacementFn,
    box_size, float,
    f: TextIO,
    axis=None,
    dr_threshold: float=0.5,
    fractional_coordinates=True,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, Callable[[Array, NeighborList], Array]]:
  """Convenience wrapper to compute :ref:`EAM energy <eam-pot>` 
  with parameters from LAMMPS using a neighbor list.."""
  return eam_neighbor_list(displacement,
                           box_size,
                           *load_lammps_eam_parameters(f),
                           dr_threshold=dr_threshold,
                           **neighbor_kwargs)


def behler_parrinello(displacement: DisplacementFn,
                      species: Optional[Array]=None,
                      mlp_sizes: Tuple[int, ...]=(30, 30),
                      mlp_kwargs: Optional[Dict[str, Any]]=None,
                      sym_kwargs: Optional[Dict[str, Any]]=None,
                      per_particle: bool=False
                      ) -> Tuple[nn.InitFn,
                                 Callable[[PyTree, Array], Array]]:
  if sym_kwargs is None:
    sym_kwargs = {}
  if mlp_kwargs is None:
    mlp_kwargs = {
        'activation': jnp.tanh
    }

  sym_fn = nn.behler_parrinello_symmetry_functions(displacement,
                                                   species,
                                                   **sym_kwargs)

  @hk.without_apply_rng
  @hk.transform
  def model(R, **kwargs):
    embedding_fn = hk.nets.MLP(output_sizes=mlp_sizes+(1,),
                               activate_final=False,
                               name='BPEncoder',
                               **mlp_kwargs)
    embedding_fn = vmap(embedding_fn)
    sym = sym_fn(R, **kwargs)
    readout = embedding_fn(sym)
    if per_particle:
      return readout
    return jnp.sum(readout)
  return model.init, model.apply


def behler_parrinello_neighbor_list(
    displacement: DisplacementFn,
    box_size: float,
    species: Optional[Array]=None,
    mlp_sizes: Tuple[int, ...]=(30, 30),
    mlp_kwargs: Optional[Dict[str, Any]]=None,
    sym_kwargs: Optional[Dict[str, Any]]=None,
    dr_threshold: float=0.5,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.Sparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, nn.InitFn, Callable[[PyTree, Array, NeighborList],
                                               Array]]:
  if sym_kwargs is None:
    sym_kwargs = {}
  if mlp_kwargs is None:
    mlp_kwargs = {
        'activation': jnp.tanh
    }

  cutoff_distance = 8.0
  if 'cutoff_distance' in sym_kwargs:
    cutoff_distance = sym_kwargs['cutoff_distance']

  neighbor_fn = partition.neighbor_list(
    displacement,
    box_size,
    cutoff_distance,
    dr_threshold,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)

  sym_fn = nn.behler_parrinello_symmetry_functions_neighbor_list(displacement,
                                                                 species,
                                                                 **sym_kwargs)

  @hk.without_apply_rng
  @hk.transform
  def model(R, neighbor, **kwargs):
    embedding_fn = hk.nets.MLP(output_sizes=mlp_sizes+(1,),
                               activate_final=False,
                               name='BPEncoder',
                               **mlp_kwargs)
    embedding_fn = vmap(embedding_fn)
    sym = sym_fn(R, neighbor, **kwargs)
    readout = embedding_fn(sym)
    return jnp.sum(readout)
  return neighbor_fn, model.init, model.apply


class EnergyGraphNet(hk.Module):
  """Implements a Graph Neural Network for energy fitting.

  This model uses a GraphNetEmbedding combined with a decoder applied to the
  global state.
  """
  def __init__(self,
               n_recurrences: int,
               mlp_sizes: Tuple[int, ...],
               mlp_kwargs: Optional[Dict[str, Any]]=None,
               format: partition.NeighborListFormat=partition.Dense,
               name: str='Energy'):
    super(EnergyGraphNet, self).__init__(name=name)

    if mlp_kwargs is None:
      mlp_kwargs = {
        'w_init': hk.initializers.VarianceScaling(),
        'b_init': hk.initializers.VarianceScaling(0.1),
        'activation': jax.nn.softplus
      }
    self._format = format
    self._graph_net = nn.GraphNetEncoder(n_recurrences,
                                         mlp_sizes,
                                         mlp_kwargs,
                                         format)
    self._decoder = hk.nets.MLP(output_sizes=mlp_sizes + (1,),
                                activate_final=False,
                                name='GlobalDecoder',
                                **mlp_kwargs)

  def __call__(self, graph: nn.GraphsTuple) -> jnp.ndarray:
    output = self._graph_net(graph)
    output = jnp.squeeze(self._decoder(output.globals), axis=-1)
    if self._format is partition.Sparse:
      output = output[0]
    return output


def _canonicalize_node_state(nodes: Optional[Array]) -> Optional[Array]:
  if nodes is None:
    return nodes

  if nodes.ndim == 1:
    nodes = nodes[:, jnp.newaxis]

  if nodes.ndim != 2:
    raise ValueError(
      'Nodes must be a [N, node_dim] array. Found {}.'.format(nodes.shape))

  return nodes


def graph_network(displacement_fn: DisplacementFn,
                  r_cutoff: float,
                  nodes: Optional[Array]=None,
                  n_recurrences: int=2,
                  mlp_sizes: Tuple[int, ...]=(64, 64),
                  mlp_kwargs: Optional[Dict[str, Any]]=None
                  ) -> Tuple[nn.InitFn,
                             Callable[[PyTree, Array], Array]]:
  """Convenience wrapper around EnergyGraphNet model.

  Args:
    displacement_fn: Function to compute displacement between two positions.
    r_cutoff: A floating point cutoff; Edges will be added to the graph
      for pairs of particles whose separation is smaller than the cutoff.
    nodes: None or an ndarray of shape `[N, node_dim]` specifying the state
      of the nodes. If None this is set to the zeros vector. Often, for a
      system with multiple species, this could be the species id.
    n_recurrences: The number of steps of message passing in the graph network.
    mlp_sizes: A tuple specifying the layer-widths for the fully-connected
      networks used to update the states in the graph network.
    mlp_kwargs: A dict specifying args for the fully-connected networks used to
      update the states in the graph network.

  Returns:
    A tuple of functions. An `params = init_fn(key, R)` that instantiates the
    model parameters and an `E = apply_fn(params, R)` that computes the energy
    for a particular state.
  """

  nodes = _canonicalize_node_state(nodes)

  @hk.without_apply_rng
  @hk.transform
  def model(R: Array, **kwargs) -> Array:
    N = R.shape[0]

    d = partial(displacement_fn, **kwargs)
    d = space.map_product(d)
    dR = d(R, R)

    dr_2 = space.square_distance(dR)

    if 'nodes' in kwargs:
      _nodes = _canonicalize_node_state(kwargs['nodes'])
    else:
      _nodes = jnp.zeros((N, 1), R.dtype) if nodes is None else nodes

    edge_idx = jnp.broadcast_to(jnp.arange(N)[jnp.newaxis, :], (N, N))
    edge_idx = jnp.where(dr_2 < r_cutoff ** 2, edge_idx, N)

    _globals = jnp.zeros((1,), R.dtype)

    net = EnergyGraphNet(n_recurrences, mlp_sizes, mlp_kwargs)
    return net(nn.GraphsTuple(_nodes, dR, _globals, edge_idx))  # pytype: disable=wrong-arg-count

  return model.init, model.apply


def graph_network_neighbor_list(
    displacement_fn: DisplacementFn,
    box_size: Box,
    r_cutoff: float,
    dr_threshold: float,
    nodes: Optional[Array]=None,
    n_recurrences: int=2,
    mlp_sizes: Tuple[int, ...]=(64, 64),
    mlp_kwargs: Optional[Dict[str, Any]]=None,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.Sparse,
    **neighbor_kwargs
    ) -> Tuple[NeighborFn, nn.InitFn, Callable[[PyTree, Array, NeighborList],
                                               Array]]:
  """Convenience wrapper around EnergyGraphNet model using neighbor lists.

  Args:
    displacement_fn: Function to compute displacement between two positions.
    box_size: The size of the simulation volume, used to construct neighbor
      list.
    r_cutoff: A floating point cutoff; Edges will be added to the graph
      for pairs of particles whose separation is smaller than the cutoff.
    dr_threshold: A floating point number specifying a "halo" radius that we use
      for neighbor list construction. See `neighbor_list` for details.
    nodes: None or an ndarray of shape `[N, node_dim]` specifying the state
      of the nodes. If None this is set to the zeroes vector. Often, for a
      system with multiple species, this could be the species id.
    n_recurrences: The number of steps of message passing in the graph network.
    mlp_sizes: A tuple specifying the layer-widths for the fully-connected
      networks used to update the states in the graph network.
    mlp_kwargs: A dict specifying args for the fully-connected networks used to
      update the states in the graph network.
    fractional_coordinates: A boolean specifying whether or not the coordinates
      will be in the unit cube.
    format: The format of the neighbor list. See `partition.NeighborListFormat`
      for details. Only `Dense` and `Sparse` formats are accepted. If the `Dense`
      format is used, then the graph network is constructed using the JAX MD
      backend, otherwise Jraph is used.

  Returns:
    A pair of functions. An `params = init_fn(key, R)` that instantiates the
    model parameters and an `E = apply_fn(params, R)` that computes the energy
    for a particular state.
  """

  nodes = _canonicalize_node_state(nodes)

  @hk.without_apply_rng
  @hk.transform
  def model(R, neighbor, **kwargs):
    N = R.shape[0]
    d = partial(displacement_fn, **kwargs)

    if 'nodes' in kwargs:
      _nodes = _canonicalize_node_state(kwargs['nodes'])
    else:
      _nodes = jnp.zeros((N, 1), R.dtype) if nodes is None else nodes

    _globals = jnp.zeros((1,), R.dtype)

    if format is partition.Dense:
      d = space.map_neighbor(d)
      R_neigh = R[neighbor.idx]
      dR = d(R, R_neigh)

      dr_2 = space.square_distance(dR)
      edge_idx = jnp.where(dr_2 < r_cutoff ** 2, neighbor.idx, N)
      graph = nn.GraphsTuple(_nodes, dR, _globals, edge_idx)
    else:
      d = space.map_bond(d)
      dR = d(R[neighbor.idx[0]], R[neighbor.idx[1]])
      if dr_threshold > 0.0:
        dr_2 = space.square_distance(dR)
        mask = dr_2 < r_cutoff ** 2 + 1e-5
        graph = partition.to_jraph(neighbor, mask)
        # TODO(schsam): It seems wasteful to recompute dR after we remask the
        # edges. If I can think of a clean way to get rid of this, I should.
        dR = d(R[graph.receivers], R[graph.senders])
      else:
        graph = partition.to_jraph(neighbor)

      graph = graph._replace(
        nodes=jnp.concatenate((_nodes,
                               jnp.zeros((1,) + _nodes.shape[1:], R.dtype)),
                             axis=0),
        edges=dR,
        globals=jnp.broadcast_to(_globals[:, None], (2, 1))
      )

    net = EnergyGraphNet(n_recurrences, mlp_sizes, mlp_kwargs, format)
    return net(graph)  # pytype: disable=wrong-arg-count

  neighbor_fn = partition.neighbor_list(
    displacement_fn,
    box_size,
    r_cutoff,
    dr_threshold,
    mask_self=False,
    fractional_coordinates=fractional_coordinates,
    format=format,
    **neighbor_kwargs)
  init_fn, apply_fn = model.init, model.apply

  return neighbor_fn, init_fn, apply_fn



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Kernel-side code for an IPython based visualization tool."""

import base64

from google.colab import output

import IPython

import jax.numpy as jnp
from jax_md import dataclasses

import json

import numpy as onp


# INTERNAL_FILE_IMPORT


renderer_code = IPython.display.HTML(
  url=('https://raw.githubusercontent.com/google/jax-md/master/'
       'jax_md/colab_tools/visualization.html')
)

SIMULATION_IDX = 0


@dataclasses.dataclass
class Disk:
  """Disk geometry elements.

  Args:
    position: An array of shape `(steps, count, dim)` or `(count, dim)`
      specifying possibly time varying positions. Here `dim` is the spatial
      dimension.
    size: An array of shape (steps, count)`, `(count,)`, or `()` specifying
      possibly time-varying / per-disk diameters.
    color: An array of shape `(steps, count, 3)` or `(count,)` specifying
      possibly time-varying / per-disk RGB colors.
    count: The number of disks.
  """
  position: jnp.ndarray
  size: jnp.ndarray
  color: jnp.ndarray
  count: int = dataclasses.static_field()

  def __init__(self, position, diameter=1.0, color=None):
    if color is None:
      color = jnp.array([0.8, 0.8, 1.0])

    object.__setattr__(self, 'position', position)
    object.__setattr__(self, 'size', diameter)
    object.__setattr__(self, 'color', color)
    object.__setattr__(self, 'count', position.shape[-2])

  def __repr__(self):
    return 'Disk'


@dataclasses.dataclass
class Sphere:
  """Sphere geometry elements.

  Args:
    position: An array of shape `(steps, count, dim)` or `(count, dim)`
      specifying possibly time varying positions. Here `dim` is the spatial
      dimension.
    size: An array of shape (steps, count)`, `(count,)`, or `()` specifying
      possibly time-varying / per-sphere diameters.
    color: An array of shape `(steps, count, 3)` or `(count,)` specifying
      possibly time-varying / per-sphere RGB colors.
    count: The number of spheres.
  """
  position: jnp.ndarray
  size: jnp.ndarray
  color: jnp.ndarray
  count: int

  def __init__(self, position, diameter=1.0, color=None):
    if color is None:
      color = jnp.array([0.8, 0.8, 1.0])

    object.__setattr__(self, 'position', position)
    object.__setattr__(self, 'size', diameter)
    object.__setattr__(self, 'color', color)
    object.__setattr__(self, 'count', position.shape[-2])

  def __repr__(self):
    return 'Sphere'


@dataclasses.dataclass
class Bond:
  """Bonds are lines between geometric objects.

  Args:
    reference_geometry: The name of the geometry object to draw bonds between.
    neighbor_idx: An array of ids of objects that should have bonds drawn
      between them. This uses the same encoding as in `partition.neighbor_list`.
      Essentially, ids is a `(steps, count, max_neighbors)` or
      `(count, max_neighbors)` array of integers. `neighbor_idx[i, j] < count`
      denotes a bond between object `i` and `neighbor[i, j]`.
    diameter: The width of the line between the objects.
    color: An array of shape `(3,)` specifying the RGB color of the bonds.
    count: The number of objects.
    max_neighbors: The maximum number of bonds a central object can have.
  """
  reference_geometry: str
  neighbor_idx: jnp.ndarray
  diameter: jnp.ndarray
  color: jnp.ndarray
  count: int
  max_neighbors: int

  def __init__(self, reference_geometry, idx, diameter=None, color=None):
    if color is None:
      color = jnp.array([0.8, 0.8, 1.0])
    if diameter is None:
      diameter = jnp.array(0.2)

    object.__setattr__(self, 'reference_geometry', reference_geometry)
    object.__setattr__(self, 'neighbor_idx', idx)
    object.__setattr__(self, 'diameter', diameter)
    object.__setattr__(self, 'color', color)
    object.__setattr__(self, 'count', idx.shape[-2])
    object.__setattr__(self, 'max_neighbors', idx.shape[-1])

  def __repr__(self):
    return 'Bond'


TYPE_DIMENSIONS = {
    'position': 2,
    'size': 1,
    'color': 2,
    'neighbor_idx': 2,
    'diameter': 1,
}


def _encode(R):
  dtype = R.dtype
  if dtype == jnp.float64:
    dtype = jnp.float32
  if dtype == jnp.int64:
    dtype = jnp.int32
  dtype = jnp.float32
  return base64.b64encode(onp.array(R, dtype).tobytes()).decode('utf-8')

def _to_json(data):
  try:
    return IPython.display.JSON(data=data)
  except:
    return IPython.display.JSON(data=json.dumps(data))

def render(box_size,
           geometry,
           buffer_size=None,
           background_color=None,
           resolution=None,
           frame_rate=None):
  """Creates a rendering front-end along with callbacks in the host program.

  Args:
    box_size: A float or an array of shape `(spatial_dimension,)`. Specifies
      the size of the simulation volume. Used to position the camera.
    geometry: A dictionary containing names paired with geometric objects such
      as Disk, Sphere, or Bond.
    buffer_size: The maximum number of timesteps to send to the font-end in a
      single call.
    background_color: An array of shape (3,) specifying the background color of
      the visualization.
    resolution: The resolution of the renderer.
    frame_rate: An optional integer specifying the target frames-per-second
      for the renderer.
  """
  global SIMULATION_IDX
  # INTERNAL_RENDERER_CODE_LOADING

  simulation_idx = SIMULATION_IDX

  frame_count = None
  dimension = None

  if not isinstance(geometry, dict):
    geometry = { 'all': geometry }
 
  for geom in geometry.values():
    if hasattr(geom, 'position'):
      assert dimension is None or geom.position.shape[-1] == dimension
      dimension = geom.position.shape[-1]

      if geom.position.ndim == 3:
        assert frame_count is None or frame_count == geom.position.shape[0]
        frame_count = geom.position.shape[0]

  assert dimension is not None

  if isinstance(box_size, jnp.ndarray):
    if box_size.shape:
      assert box_size.shape == (dimension,)
      box_size = list(box_size)
    else:
      box_size = [float(box_size),] * dimension
  elif isinstance(box_size, float) or isinstance(box_size, int):
    box_size = [box_size,] * dimension

  def get_metadata():
    metadata = {
        'box_size': box_size,
        'dimension': dimension,
        'geometry': [k for k in geometry.keys()],
        'simulation_idx': simulation_idx
    }

    if frame_count is not None:
      metadata['frame_count'] = frame_count

    if buffer_size is not None:
      metadata['buffer_size'] = buffer_size

    if background_color is not None:
      metadata['background_color'] = background_color

    if resolution is not None:
      metadata['resolution'] = resolution

    if frame_rate is not None:
      metadata['frame_rate'] = frame_rate

    return _to_json(metadata)
  output.register_callback('GetSimulationMetadata', get_metadata)

  def get_dynamic_geometry_metadata(name):
    assert name in geometry

    geom = geometry[name]
    geom_dict = dataclasses.asdict(geom)

    geom_metadata = {
        'shape': str(geom),
        'fields': {},
    }

    for field in geom_dict:
      if not isinstance(geom_dict[field], onp.ndarray):
        geom_metadata[field] = geom_dict[field]
        continue
      if len(geom_dict[field].shape) == TYPE_DIMENSIONS[field] + 1:
        geom_metadata['fields'][field] = 'dynamic'
      elif len(geom_dict[field].shape) == TYPE_DIMENSIONS[field]:
        geom_metadata['fields'][field] = 'static'
      elif len(geom_dict[field].shape) == TYPE_DIMENSIONS[field] - 1:
        geom_metadata['fields'][field] = 'global'
    return _to_json(geom_metadata)
  output.register_callback(f'GetGeometryMetadata{SIMULATION_IDX}',
                          get_dynamic_geometry_metadata)

  def get_array_chunk(name, field, offset, size):
    assert name in geometry

    geom = dataclasses.asdict(geometry[name])
    assert field in geom
    array = geom[field]

    return _to_json({
        'array_chunk': _encode(array[offset:(offset + size)])
    })
  output.register_callback(f'GetArrayChunk{SIMULATION_IDX}', get_array_chunk)

  def get_array(name, field):
    assert name in geometry

    geom = dataclasses.asdict(geometry[name])
    assert field in geom
    array = geom[field]

    return _to_json({ 'array': _encode(array) })
  output.register_callback(f'GetArray{SIMULATION_IDX}', get_array)

  SIMULATION_IDX += 1

  IPython.display.display(renderer_code)



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import io
import os
import setuptools

# https://packaging.python.org/guides/making-a-pypi-friendly-readme/
this_directory = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
  long_description = f.read()

INSTALL_REQUIRES = [
    'absl-py',
    'numpy',
    'jax>=0.1.73',
    'jaxlib>=0.1.52',
    'dm-haiku>=0.0.2',
    'jraph',
    'dataclasses'
]

setuptools.setup(
    name='jax-md',
    version='0.1.27',
    license='Apache 2.0',
    author='Google',
    author_email='jax-md-dev@google.com',
    install_requires=INSTALL_REQUIRES,
    url='https://github.com/google/jax-md',
    packages=setuptools.find_packages(),
    download_url = "https://pypi.org/project/jax-md/",
    project_urls={
        "Source Code": "https://github.com/google/jax-md",
        "Documentation": "https://arxiv.org/abs/1912.04232",
        "Bug Tracker": "https://github.com/google/jax-md/issues",
    },
    long_description=long_description,
    long_description_content_type='text/markdown',
    description='Differentiable, Hardware Accelerated, Molecular Dynamics',
    python_requires='>=2.7',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: MacOS',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development',
        'Topic :: Scientific/Engineering',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
    ])



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jax.numpy as np

def _xyz_vec2str(x):
  """Convert a vector to string.
  """
  return "\t".join([str(i) for i in x])

def write_xyz(filename, *args):
  """Write arrays to xyz file format.

  Args:
    filename: Output filename.
    args: Arguments can be 1D or 2D arrays where length (along axis=0) is equal to number of atoms.

  Examples:
    write_xyz('minimize.xyz', R)
    write_xyz('minimize.xyz', species, R)
    write_xyz('minimize.xyz', species, R, velocities, forces)
  """
  vars = [arg[:, np.newaxis] if len(arg.shape)==1 else arg for arg in args]
  vars = np.hstack(vars)

  with open(filename, "w+") as f:
    N = len(vars)
    str_ = f"{N}" + "\n\n"
    f.write(str_)
    for j in range(N):
      str_ = f"{j+1}\t" + _xyz_vec2str(vars[j, :]) + "\n"
      f.write(str_)



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Neural Network Primitives."""

from typing import Callable, Tuple, Dict, Any, Optional

import numpy as onp

import jax
from jax import vmap, jit
import jax.numpy as jnp

from jax_md import space, dataclasses, quantity, partition, smap, util
import haiku as hk

from collections import namedtuple
from functools import partial, reduce
from jax.tree_util import tree_multimap, tree_map
from jax import ops

import jraph

# Typing


Array = util.Array
f32 = util.f32
f64 = util.f64

InitFn = Callable[..., Array]
CallFn = Callable[..., Array]

DisplacementOrMetricFn = space.DisplacementOrMetricFn
DisplacementFn = space.DisplacementFn
NeighborList = partition.NeighborList


# Features used in fixed feature methods


"""
Our neural network force field is based on the Behler-Parrinello neural network
architecture (BP-NN) [1]. The BP-NN architecture uses a relatively simple,
fully connected neural network to predict the local energy for each atom. Then
the total energy is the sum of local energies due to each atom. Atoms of the
same type use the same NN to predict energy.

Each atomic NN is applied to hand-crafted features called symmetry functions.
There are two kinds of symmetry functions: radial and angular. Radial symmetry
functions represent information about two-body interactions of the central
atom, whereas angular symmetry functions represent information about three-body
interactions. Below we implement radial and angular symmetry functions for
arbitrary number of types of atoms (Note that most applications of BP-NN limit
their systems to 1 to 3 types of atoms). We also present a convenience wrapper
that returns radial and angular symmetry functions with symmetry function
parameters that should work reasonably for most systems (the symmetry functions
are taken from reference [2]). Please see references [1, 2] for details about
how the BP-NN works.

[1] Behler, Jörg, and Michele Parrinello. "Generalized neural-network
representation of high-dimensional potential-energy surfaces." Physical
Review Letters 98.14 (2007): 146401.

[2] Artrith, Nongnuch, Björn Hiller, and Jörg Behler. "Neural network
potentials for metals and oxides–First applications to copper clusters at zinc
oxide."
Physica Status Solidi (b) 250.6 (2013): 1191-1203.
"""

def _behler_parrinello_cutoff_fn(dr: Array,
                                 cutoff_distance: float=8.0) -> Array:
  """Function of pairwise distance that smoothly goes to zero at the cutoff."""
  # Also returns zero if the pairwise distance is zero,
  # to prevent a particle from interacting with itself.
  return jnp.where((dr < cutoff_distance) & (dr > 1e-7),
                  0.5 * (jnp.cos(jnp.pi * dr / cutoff_distance) + 1), 0)


def radial_symmetry_functions(displacement_or_metric: DisplacementOrMetricFn,
                              species: Optional[Array],
                              etas: Array,
                              cutoff_distance: float
                              ) -> Callable[[Array], Array]:
  """Returns a function that computes radial symmetry functions.


  Args:
    displacement: A function that produces an `[N_atoms, M_atoms,
      spatial_dimension]` of particle displacements from particle positions
      specified as an `[N_atoms, spatial_dimension] and `[M_atoms,
      spatial_dimension]` respectively.
    species: An `[N_atoms]` that contains the species of each particle.
    etas: List of radial symmetry function parameters that control the spatial
      extension.
    cutoff_distance: Neighbors whose distance is larger than `cutoff_distance` do
      not contribute to each others symmetry functions. The contribution of a
      neighbor to the symmetry function and its derivative goes to zero at this
      distance.

  Returns:
    A function that computes the radial symmetry function from input `[N_atoms,
    spatial_dimension]` and returns `[N_atoms, N_etas * N_types]` where `N_etas`
    is the number of eta parameters, `N_types` is the number of types of
    particles in the system.
  """
  metric = space.canonicalize_displacement_or_metric(displacement_or_metric)

  radial_fn = lambda eta, dr: (jnp.exp(-eta * dr**2) *
                               _behler_parrinello_cutoff_fn(dr, cutoff_distance))
  radial_fn = vmap(radial_fn, (0, None))

  if species is None:
    def compute_fn(R: Array, **kwargs) -> Array:
      _metric = partial(metric, **kwargs)
      _metric = space.map_product(_metric)
      return util.high_precision_sum(radial_fn(etas, _metric(R, R)), axis=1).T
  elif isinstance(species, jnp.ndarray):
    species = onp.array(species)
    def compute_fn(R: Array, **kwargs) -> Array:
      _metric = partial(metric, **kwargs)
      _metric = space.map_product(_metric)
      def return_radial(atom_type):
        """Returns the radial symmetry functions for neighbor type atom_type."""
        R_neigh = R[species == atom_type, :]
        dr = _metric(R, R_neigh)
        return util.high_precision_sum(radial_fn(etas, dr), axis=1).T
      return jnp.hstack([return_radial(atom_type) for
                        atom_type in onp.unique(species)])
  return compute_fn


def radial_symmetry_functions_neighbor_list(
    displacement_or_metric: DisplacementOrMetricFn,
    species: Array,
    etas: Array,
    cutoff_distance: float) -> Callable[[Array, NeighborList], Array]:
  """Returns a function that computes radial symmetry functions.


  Args:
    displacement: A function that produces an `[N_atoms, M_atoms,
      spatial_dimension]` of particle displacements from particle positions
      specified as an `[N_atoms, spatial_dimension] and `[M_atoms,
      spatial_dimension]` respectively.
    species: An `[N_atoms]` that contains the species of each particle.
    etas: List of radial symmetry function parameters that control the spatial
      extension.
    cutoff_distance: Neighbors whose distance is larger than `cutoff_distance` do
      not contribute to each others symmetry functions. The contribution of a
      neighbor to the symmetry function and its derivative goes to zero at this
      distance.

  Returns:
    A function that computes the radial symmetry function from input `[N_atoms,
    spatial_dimension]` and returns `[N_atoms, N_etas * N_types]` where `N_etas`
    is the number of eta parameters, `N_types` is the number of types of
    particles in the system.
  """
  metric = space.canonicalize_displacement_or_metric(displacement_or_metric)

  def radial_fn(eta, dr):
    return (jnp.exp(-eta * dr**2) *
            _behler_parrinello_cutoff_fn(dr, cutoff_distance))
  radial_fn = vmap(radial_fn, (0, None))

  def sym_fn(R: Array, neighbor: NeighborList, mask: Array=None,
             **kwargs) -> Array:
    _metric = partial(metric, **kwargs)
    if neighbor.format is partition.Dense:
      _metric = space.map_neighbor(_metric)
      R_neigh = R[neighbor.idx]
      mask = True if mask is None else mask[neighbor.idx]
      mask = (neighbor.idx < R.shape[0])[None, :, :] & mask
      dr = _metric(R, R_neigh)
      return util.high_precision_sum(radial_fn(etas, dr) * mask, axis=2).T
    elif neighbor.format is partition.Sparse:
      _metric = space.map_bond(_metric)
      dr = _metric(R[neighbor.idx[0]], R[neighbor.idx[1]])
      radial = radial_fn(etas, dr).T
      N = R.shape[0]
      mask = True if mask is None else mask[neighbor.idx[1]]
      mask = (neighbor.idx[0] < N) & mask
      return ops.segment_sum(radial * mask[:, None], neighbor.idx[0], N)
    else:
      raise ValueError()

  if species is None:
    def compute_fn(R: Array, neighbor: NeighborList, **kwargs) -> Array:
      return sym_fn(R, neighbor, **kwargs)
    return compute_fn

  def compute_fn(R: Array, neighbor: NeighborList, **kwargs) -> Array:
    _metric = partial(metric, **kwargs)
    def return_radial(atom_type):
      """Returns the radial symmetry functions for neighbor type atom_type."""
      return sym_fn(R, neighbor, species==atom_type, **kwargs)
    return jnp.hstack([return_radial(atom_type) for
                     atom_type in onp.unique(species)])

  return compute_fn


def single_pair_angular_symmetry_function(dR12: Array,
                                          dR13: Array,
                                          eta: Array,
                                          lam: Array,
                                          zeta: Array,
                                          cutoff_distance: float
                                          ) -> Array:
  """Computes the angular symmetry function due to one pair of neighbors."""

  dR23 = dR12 - dR13
  dr12_2 = space.square_distance(dR12)
  dr13_2 = space.square_distance(dR13)
  dr23_2 = space.square_distance(dR23)
  dr12 = space.distance(dR12)
  dr13 = space.distance(dR13)
  dr23 = space.distance(dR23)
  triplet_squared_distances = dr12_2 + dr13_2 + dr23_2
  triplet_cutoff = reduce(
      lambda x, y: x * _behler_parrinello_cutoff_fn(y, cutoff_distance),
      [dr12, dr13, dr23], 1.0)
  z = zeta
  result = 2.0 ** (1.0 - zeta) * ((
    1.0 + lam * quantity.cosine_angle_between_two_vectors(dR12, dR13)) ** z *
    jnp.exp(-eta * triplet_squared_distances) * triplet_cutoff)
  return result


def angular_symmetry_functions(displacement: DisplacementFn,
                               species: Array,
                               etas: Array,
                               lambdas: Array,
                               zetas: Array,
                               cutoff_distance: float
                               ) -> Callable[[Array], Array]:
  """Returns a function that computes angular symmetry functions.

  Args:
    displacement: A function that produces an `[N_atoms, M_atoms,
      spatial_dimension]` of particle displacements from particle positions
      specified as an `[N_atoms, spatial_dimension] and `[M_atoms,
      spatial_dimension]` respectively.
    species: An `[N_atoms]` that contains the species of each particle.
    eta: Parameter of angular symmetry function that controls the spatial
      extension.
    lam:
    zeta:
    cutoff_distance: Neighbors whose distance is larger than `cutoff_distance` do
      not contribute to each others symmetry functions. The contribution of a
      neighbor to the symmetry function and its derivative goes to zero at this
      distance.
  Returns:
    A function that computes the angular symmetry function from input `[N_atoms,
    spatial_dimension]` and returns `[N_atoms, N_types * (N_types + 1) / 2]`
    where `N_types` is the number of types of particles in the system.
  """

  _angular_fn = vmap(single_pair_angular_symmetry_function,
                     (None, None, 0, 0, 0, None))

  _batched_angular_fn = lambda dR12, dR13: _angular_fn(dR12,
                                                       dR13,
                                                       etas,
                                                       lambdas,
                                                       zetas,
                                                       cutoff_distance)
  _all_pairs_angular = vmap(
      vmap(vmap(_batched_angular_fn, (0, None)), (None, 0)), 0)

  if species is None:
   def compute_fn(R, **kwargs):
     D_fn = partial(displacement, **kwargs)
     D_fn = space.map_product(D_fn)
     dR = D_fn(R, R)
     return jnp.sum(_all_pairs_angular(dR, dR), axis=[1, 2])
   return compute_fn

  if isinstance(species, jnp.ndarray):
    species = onp.array(species)

  def compute_fn(R, **kwargs):
    atom_types = onp.unique(species)
    D_fn = partial(displacement, **kwargs)
    D_fn = space.map_product(D_fn)
    D_different_types = [D_fn(R[species == s, :], R) for s in atom_types]
    out = []
    for i in range(len(atom_types)):
      for j in range(i, len(atom_types)):
        out += [
            jnp.sum(
                _all_pairs_angular(D_different_types[i], D_different_types[j]),
                axis=[1, 2])
        ]
    return jnp.hstack(out)
  return compute_fn

def angular_symmetry_functions_neighbor_list(
    displacement: DisplacementFn,
    species: Array,
    etas: Array,
    lambdas: Array,
    zetas: Array,
    cutoff_distance: float) -> Callable[[Array, NeighborList], Array]:
  """Returns a function that computes angular symmetry functions.

  Args:
    displacement: A function that produces an `[N_atoms, M_atoms,
      spatial_dimension]` of particle displacements from particle positions
      specified as an `[N_atoms, spatial_dimension] and `[M_atoms,
      spatial_dimension]` respectively.
    species: An `[N_atoms]` that contains the species of each particle.
    eta: Parameter of angular symmetry function that controls the spatial
      extension.
    lam:
    zeta:
    cutoff_distance: Neighbors whose distance is larger than `cutoff_distance` do
      not contribute to each others symmetry functions. The contribution of a
      neighbor to the symmetry function and its derivative goes to zero at this
      distance.
  Returns:
    A function that computes the angular symmetry function from input
    `[N_atoms, spatial_dimension]` and returns
    `[N_atoms, N_types * (N_types + 1) / 2]` where `N_types` is the number of
    types of particles in the system.
  """

  _angular_fn = vmap(single_pair_angular_symmetry_function,
                     (None, None, 0, 0, 0, None))

  _batched_angular_fn = lambda dR12, dR13: _angular_fn(dR12,
                                                       dR13,
                                                       etas,
                                                       lambdas,
                                                       zetas,
                                                       cutoff_distance)
  _all_pairs_angular = vmap(
      vmap(vmap(_batched_angular_fn, (0, None)), (None, 0)), 0)

  def sym_fn(R: Array, neighbor: NeighborList,
             mask_i: Array=None, mask_j: Array=None,
             **kwargs) -> Array:
    D_fn = partial(displacement, **kwargs)

    if neighbor.format is partition.Dense:
      D_fn = space.map_neighbor(D_fn)

      R_neigh = R[neighbor.idx]

      dR = D_fn(R, R_neigh)
      _all_pairs_angular = vmap(
        vmap(vmap(_batched_angular_fn, (0, None)), (None, 0)), 0)
      all_angular = _all_pairs_angular(dR, dR)

      mask_i = True if mask_i is None else mask_i[neighbor.idx]
      mask_j = True if mask_j is None else mask_j[neighbor.idx]

      mask_i = (neighbor.idx < R.shape[0]) & mask_i
      mask_i = mask_i[:, :, jnp.newaxis, jnp.newaxis]
      mask_j = (neighbor.idx < R.shape[0]) & mask_j
      mask_j = mask_j[:, jnp.newaxis, :, jnp.newaxis]

      return util.high_precision_sum(all_angular * mask_i * mask_j,
                                     axis=[1, 2])
    elif neighbor.format is partition.Sparse:
      D_fn = space.map_bond(D_fn)
      dR = D_fn(R[neighbor.idx[0]], R[neighbor.idx[1]])
      _all_pairs_angular = vmap(vmap(_batched_angular_fn, (0, None)),
                                (None, 0))
      all_angular = _all_pairs_angular(dR, dR)

      N = R.shape[0]
      mask_i = True if mask_i is None else mask_i[neighbor.idx[1]]
      mask_j = True if mask_j is None else mask_j[neighbor.idx[1]]
      mask_i = (neighbor.idx[0] < N) & mask_i
      mask_j = (neighbor.idx[0] < N) & mask_j

      mask = mask_i[:, None] & mask_j[None, :]
      mask = mask[:, :, None, None]
      all_angular = jnp.reshape(all_angular, (-1,) + all_angular.shape[2:])
      neighbor_idx = jnp.repeat(neighbor.idx[0], len(neighbor.idx[0]))
      out = ops.segment_sum(all_angular, neighbor_idx, N)
      return out
    else:
      raise ValueError()

  if species is None:
    def compute_fn(R: Array, neighbor: NeighborList, **kwargs) -> Array:
      return sym_fn(R, neighbor, **kwargs)
    return compute_fn

  def compute_fn(R: Array, neighbor: NeighborList, **kwargs) -> Array:
    atom_types = onp.unique(species)
    out = []
    for i in range(len(atom_types)):
      mask_i = species == i
      for j in range(i, len(atom_types)):
        mask_j = species == j
        out += [
            sym_fn(R, neighbor, mask_i, mask_j)
        ]
    return jnp.hstack(out)
  return compute_fn


def behler_parrinello_symmetry_functions_neighbor_list(
    displacement: DisplacementFn,
    species: Array,
    radial_etas: Optional[Array]=None,
    angular_etas: Optional[Array]=None,
    lambdas: Optional[Array]=None,
    zetas: Optional[Array]=None,
    cutoff_distance: float=8.0) -> Callable[[Array, NeighborList], Array]:
  if radial_etas is None:
    radial_etas = jnp.array([9e-4, 0.01, 0.02, 0.035, 0.06, 0.1, 0.2, 0.4],
                    f32) / f32(0.529177 ** 2)

  if angular_etas is None:
    angular_etas = jnp.array([1e-4] * 4 + [0.003] * 4 + [0.008] * 2 +
                            [0.015] * 4 + [0.025] * 4 + [0.045] * 4,
                            f32) / f32(0.529177 ** 2)

  if lambdas is None:
    lambdas = jnp.array([-1, 1] * 4 + [1] * 14, f32)

  if zetas is None:
    zetas = jnp.array([1, 1, 2, 2] * 2 + [1, 2] + [1, 2, 4, 16] * 3, f32)

  radial_fn = radial_symmetry_functions_neighbor_list(
    displacement,
    species,
    etas=radial_etas,
    cutoff_distance=cutoff_distance)
  angular_fn = angular_symmetry_functions_neighbor_list(
    displacement,
    species,
    etas=angular_etas,
    lambdas=lambdas,
    zetas=zetas,
    cutoff_distance=cutoff_distance)
  return (lambda R, neighbor, **kwargs:
          jnp.hstack((radial_fn(R, neighbor, **kwargs),
                     angular_fn(R, neighbor, **kwargs))))


def behler_parrinello_symmetry_functions(displacement: DisplacementFn,
                                         species: Optional[Array]=None,
                                         radial_etas: Optional[Array]=None,
                                         angular_etas: Optional[Array]=None,
                                         lambdas: Optional[Array]=None,
                                         zetas: Optional[Array]=None,
                                         cutoff_distance: float=8.0
                                         ) -> Callable[[Array], Array]:
  if radial_etas is None:
    radial_etas = jnp.array([9e-4, 0.01, 0.02, 0.035, 0.06, 0.1, 0.2, 0.4],
                    f32) / f32(0.529177 ** 2)

  if angular_etas is None:
    angular_etas = jnp.array([1e-4] * 4 + [0.003] * 4 + [0.008] * 2 +
                            [0.015] * 4 + [0.025] * 4 + [0.045] * 4,
                            f32) / f32(0.529177 ** 2)

  if lambdas is None:
    lambdas = jnp.array([-1, 1] * 4 + [1] * 14, f32)

  if zetas is None:
    zetas = jnp.array([1, 1, 2, 2] * 2 + [1, 2] + [1, 2, 4, 16] * 3, f32)

  radial_fn = radial_symmetry_functions(displacement,
                                        species,
                                        etas=radial_etas,
                                        cutoff_distance=cutoff_distance)
  angular_fn = angular_symmetry_functions(displacement,
                                          species,
                                          etas=angular_etas,
                                          lambdas=lambdas,
                                          zetas=zetas,
                                          cutoff_distance=cutoff_distance)
  symmetry_fn = lambda R, **kwargs: jnp.hstack((radial_fn(R, **kwargs),
                                               angular_fn(R, **kwargs)))

  return symmetry_fn


# Graph neural network primitives

"""
  Our implementation here is based off the outstanding GraphNets library by
  DeepMind at, www.github.com/deepmind/graph_nets. This implementation was also
  heavily influenced by work done by Thomas Keck. We implement a subset of the
  functionality from the graph nets library to be compatible with jax-md
  states and neighbor lists, end-to-end jit compilation, and easy batching.

  Graphs are described by node states, edge states, a global state, and
  outgoing / incoming edges.

  We provide two components:

    1) A GraphIndependent layer that applies a neural network separately to the
       node states, the edge states, and the globals. This is often used as an
       encoding or decoding step.
    2) A GraphNetwork layer that transforms the nodes, edges, and globals using
       neural networks following Battaglia et al. (). Here, we use
       sum-message-aggregation.

  The graphs network components implemented here implement identical functions
  to the DeepMind library. However, to be compatible with jax-md, there are
  significant differences in the graph layout used here to the reference
  implementation. See `GraphsTuple` for details.
"""

@dataclasses.dataclass
class GraphsTuple(object):
    """A struct containing graph data.

    Attributes:
      nodes: For a graph with `N_nodes`, this is an `[N_nodes, node_dimension]`
        array containing the state of each node in the graph.
      edges: For a graph whose degree is bounded by max_degree, this is an
        `[N_nodes, max_degree, edge_dimension]`. Here `edges[i, j]` is the
        state of the outgoing edge from node `i` to node `edge_idx[i, j]`.
      globals: An array of shape `[global_dimension]`.
      edge_idx: An integer array of shape `[N_nodes, max_degree]` where
        `edge_idx[i, j]` is the id of the j-th outgoing edge from node `i`.
        Empty entries (that don't contain an edge) are denoted by
        `edge_idx[i, j] == N_nodes`.
    """
    nodes: jnp.ndarray
    edges: jnp.ndarray
    globals: jnp.ndarray
    edge_idx: jnp.ndarray

    _replace = dataclasses.replace


def concatenate_graph_features(graphs: Tuple[GraphsTuple, ...]) -> GraphsTuple:
  """Given a list of GraphsTuple returns a new concatenated GraphsTuple.

  Note that currently we do not check that the graphs have consistent edge
  connectivity.
  """
  graph = graphs[0]
  return graph._replace(
    nodes=jnp.concatenate([g.nodes for g in graphs], axis=-1),
    edges=jnp.concatenate([g.edges for g in graphs], axis=-1),
    globals=jnp.concatenate([g.globals for g in graphs], axis=-1),  # pytype: disable=missing-parameter
  )


def GraphMapFeatures(edge_fn: Callable[[Array], Array],
                     node_fn: Callable[[Array], Array],
                     global_fn: Callable[[Array], Array]
                     ) -> Callable[[GraphsTuple], GraphsTuple]:
  """Applies functions independently to the nodes, edges, and global states.
  """
  identity = lambda x: x
  _node_fn = vmap(node_fn) if node_fn is not None else identity
  _edge_fn = vmap(vmap(edge_fn)) if edge_fn is not None else identity
  _global_fn = global_fn if global_fn is not None else identity

  def embed_fn(graph):
    return dataclasses.replace(
        graph,
        nodes=_node_fn(graph.nodes),
        edges=_edge_fn(graph.edges),
        globals=_global_fn(graph.globals)
    )
  return embed_fn


def _apply_node_fn(graph: GraphsTuple,
                   node_fn: Callable[[Array,Array, Array, Array], Array]
                   ) -> Array:
  mask = graph.edge_idx < graph.nodes.shape[0]
  mask = mask[:, :, jnp.newaxis]

  if graph.edges is not None:
    # TODO: Should we also have outgoing edges?
    flat_edges = jnp.reshape(graph.edges, (-1, graph.edges.shape[-1]))
    edge_idx = jnp.reshape(graph.edge_idx, (-1,))
    incoming_edges = jax.ops.segment_sum(
        flat_edges, edge_idx, graph.nodes.shape[0] + 1)[:-1]
    outgoing_edges = jnp.sum(graph.edges * mask, axis=1)
  else:
    incoming_edges = None
    outgoing_edges = None

  if graph.globals is not None:
    _globals = jnp.broadcast_to(graph.globals[jnp.newaxis, :],
                               graph.nodes.shape[:1] + graph.globals.shape)
  else:
    _globals = None

  return node_fn(graph.nodes, incoming_edges, outgoing_edges, _globals)


def _apply_edge_fn(graph: GraphsTuple,
                   edge_fn: Callable[[Array, Array, Array, Array], Array]
                   ) -> Array:
  if graph.nodes is not None:
    incoming_nodes = graph.nodes[graph.edge_idx]
    outgoing_nodes = jnp.broadcast_to(
        graph.nodes[:, jnp.newaxis, :],
        graph.edge_idx.shape + graph.nodes.shape[-1:])
  else:
    incoming_nodes = None
    outgoing_nodes = None

  if graph.globals is not None:
    _globals = jnp.broadcast_to(graph.globals[jnp.newaxis, jnp.newaxis, :],
                               graph.edge_idx.shape + graph.globals.shape)
  else:
    _globals = None

  mask = graph.edge_idx < graph.nodes.shape[0]
  mask = mask[:, :, jnp.newaxis]
  return edge_fn(graph.edges, incoming_nodes, outgoing_nodes, _globals) * mask


def _apply_global_fn(graph: GraphsTuple,
                     global_fn: Callable[[Array, Array, Array], Array]
                     ) -> Array:
  nodes = None if graph.nodes is None else jnp.sum(graph.nodes, axis=0)

  if graph.edges is not None:
    mask = graph.edge_idx < graph.nodes.shape[0]
    mask = mask[:, :, jnp.newaxis]
    edges = jnp.sum(graph.edges * mask, axis=(0, 1))
  else:
    edges = None

  return global_fn(nodes, edges, graph.globals)


class GraphNetwork:
  """Implementation of a Graph Network.

  See https://arxiv.org/abs/1806.01261 for more details.
  """
  def __init__(self,
               edge_fn: Callable[[Array], Array],
               node_fn: Callable[[Array], Array],
               global_fn: Callable[[Array], Array]):
    self._node_fn = (None if node_fn is None else
                     partial(_apply_node_fn, node_fn=vmap(node_fn)))

    self._edge_fn = (None if edge_fn is None else
                     partial(_apply_edge_fn, edge_fn=vmap(vmap(edge_fn))))

    self._global_fn = (None if global_fn is None else
                       partial(_apply_global_fn, global_fn=global_fn))

  def __call__(self, graph: GraphsTuple) -> GraphsTuple:
    if self._edge_fn is not None:
      graph = dataclasses.replace(graph, edges=self._edge_fn(graph))

    if self._node_fn is not None:
      graph = dataclasses.replace(graph, nodes=self._node_fn(graph))

    if self._global_fn is not None:
      graph = dataclasses.replace(graph, globals=self._global_fn(graph))

    return graph


# Prefab Networks


class GraphNetEncoder(hk.Module):
  """Implements a Graph Neural Network for energy fitting.

  Based on the network used in "Unveiling the predictive power of static
  structure in glassy systems"; Bapst et al.
  (https://www.nature.com/articles/s41567-020-0842-8). This network first
  embeds edges, nodes, and global state. Then `n_recurrences` of GraphNetwork
  layers are applied. Unlike in Bapst et al. this network does not include a
  readout, which should be added separately depending on the application.

  For example, when predicting particle mobilities, one would use a decoder
  only on the node states while a model of energies would decode only the node
  states.
  """
  def __init__(self,
               n_recurrences: int,
               mlp_sizes: Tuple[int, ...],
               mlp_kwargs: Optional[Dict[str, Any]]=None,
               format: partition.NeighborListFormat=partition.Dense,
               name: str='GraphNetEncoder'):
    super(GraphNetEncoder, self).__init__(name=name)

    if mlp_kwargs is None:
      mlp_kwargs = {}

    self._n_recurrences = n_recurrences

    embedding_fn = lambda name: hk.nets.MLP(
        output_sizes=mlp_sizes,
        activate_final=True,
        name=name,
        **mlp_kwargs)

    model_fn = lambda name: lambda *args: hk.nets.MLP(
        output_sizes=mlp_sizes,
        activate_final=True,
        name=name,
        **mlp_kwargs)(jnp.concatenate(args, axis=-1))

    if format is partition.Dense:
      self._encoder = GraphMapFeatures(
        embedding_fn('EdgeEncoder'),
        embedding_fn('NodeEncoder'),
        embedding_fn('GlobalEncoder'))
      self._propagation_network = lambda: GraphNetwork(
        model_fn('EdgeFunction'),
        model_fn('NodeFunction'),
        model_fn('GlobalFunction'))
    elif format is partition.Sparse:
      self._encoder = jraph.GraphMapFeatures(
        embedding_fn('EdgeEncoder'),
        embedding_fn('NodeEncoder'),
        embedding_fn('GlobalEncoder')
      )
      self._propagation_network = lambda: jraph.GraphNetwork(
        model_fn('EdgeFunction'),
        model_fn('NodeFunction'),
        model_fn('GlobalFunction')
      )
    else:
      raise ValueError()

  def __call__(self, graph: GraphsTuple) -> GraphsTuple:
    encoded = self._encoder(graph)
    outputs = encoded

    for _ in range(self._n_recurrences):
      inputs = concatenate_graph_features((outputs, encoded))
      outputs = self._propagation_network()(inputs)

    return outputs



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Code to minimize the energy of a system.

  This file contains a number of different methods that can be used to find the
  nearest minimum (inherent structure) to some initial system described by a
  position R.

  Minimization code follows the same overall structure as optimizers in JAX.
  Optimizers return two functions:
    init_fn: 
      Function that initializes the  state of an optimizer. Should take
      positions as an ndarray of shape `[n, output_dimension]`. Returns a state
      which will be a namedtuple.
    apply_fn: 
      Function that takes a state and produces a new state after one
      step of optimization.
"""

from collections import namedtuple

from typing import TypeVar, Callable, Tuple, Union

import jax.numpy as jnp

from jax_md import quantity
from jax_md import dataclasses
from jax_md import util
from jax_md import space

# Types

Array = util.Array
f32 = util.f32
f64 = util.f64

ShiftFn = space.ShiftFn

T = TypeVar('T')
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Minimizer = Tuple[InitFn, ApplyFn]


def gradient_descent(energy_or_force: Callable[..., Array],
                     shift_fn: ShiftFn,
                     step_size: float) -> Minimizer[Array]:
  """Defines gradient descent minimization.

    This is the simplest optimization strategy that moves particles down their
    gradient to the nearest minimum. Generally, gradient descent is slower than
    other methods and is included mostly for its simplicity.

    Args:
      energy_or_force: A function that produces either an energy or a force from
        a set of particle positions specified as an ndarray of shape
        `[n, spatial_dimension]`.
      shift_fn: A function that displaces positions, `R`, by an amount `dR`. Both `R`
        and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
      step_size: A floating point specifying the size of each step.

    Returns:
      See above.
  """
  force = quantity.canonicalize_force(energy_or_force)
  def init_fn(R: Array, **unused_kwargs) -> Array:
    return R
  def apply_fn(R: Array, **kwargs) -> Array:
    R = shift_fn(R, step_size * force(R, **kwargs), **kwargs)
    return R
  return init_fn, apply_fn


@dataclasses.dataclass
class FireDescentState:
  """A dataclass containing state information for the Fire Descent minimizer.

  Attributes:
    position: The current position of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    velocity: The current velocity of particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    force: The current force on particles. An ndarray of floats
      with shape `[n, spatial_dimension]`.
    dt: A float specifying the current step size.
    alpha: A float specifying the current momentum.
    n_pos: The number of steps in the right direction, so far.
  """
  position: Array
  velocity: Array
  force: Array
  dt: float
  alpha: float
  n_pos: int


def fire_descent(energy_or_force: Callable[..., Array],
                 shift_fn: ShiftFn,
                 dt_start: float=0.1,
                 dt_max: float=0.4,
                 n_min: float=5,
                 f_inc: float=1.1,
                 f_dec: float=0.5,
                 alpha_start: float=0.1,
                 f_alpha: float=0.99) -> Minimizer[FireDescentState]:
  """Defines FIRE minimization.

  This code implements the "Fast Inertial Relaxation Engine" from Bitzek et al. [#bitzek]_

  Args:
    energy_or_force: A function that produces either an energy or a force from
      a set of particle positions specified as an ndarray of shape
      `[n, spatial_dimension]`.
    shift_fn: A function that displaces positions `R`, by an amount `dR`. Both `R`
      and `dR` should be ndarrays of shape `[n, spatial_dimension]`.
    quant: Either a quantity.Energy or a quantity. Force specifying whether
      energy_or_force is an energy or force respectively.
    dt_start: The initial step size during minimization as a float.
    dt_max: The maximum step size during minimization as a float.
    n_min: An integer specifying the minimum number of steps moving in the
      correct direction before dt and f_alpha should be updated.
    f_inc: A float specifying the fractional rate by which the step size
      should be increased.
    f_dec: A float specifying the fractional rate by which the step size
      should be decreased.
    alpha_start: A float specifying the initial momentum.
    f_alpha: A float specifying the fractional change in momentum.
  Returns:
    See above.

  .. rubric:: References
  .. [#bitzek] Bitzek, Erik, Pekka Koskinen, Franz Gahler, Michael Moseler,
      and Peter Gumbsch. "Structural relaxation made simple."
      Physical review letters 97, no. 17 (2006): 170201.
  """

  dt_start, dt_max, n_min, f_inc, f_dec, alpha_start, f_alpha = util.static_cast(
    dt_start, dt_max, n_min, f_inc, f_dec, alpha_start, f_alpha)

  force = quantity.canonicalize_force(energy_or_force)
  def init_fn(R: Array, **kwargs) -> FireDescentState:
    V = jnp.zeros_like(R)
    n_pos = jnp.zeros((), jnp.int32)
    F = force(R, **kwargs)
    return FireDescentState(R, V, F, dt_start, alpha_start, n_pos)  # pytype: disable=wrong-arg-count
  def apply_fn(state: FireDescentState, **kwargs) -> FireDescentState:
    R, V, F_old, dt, alpha, n_pos = dataclasses.astuple(state)

    R = shift_fn(R, dt * V + dt ** f32(2) * F_old, **kwargs)

    F = force(R, **kwargs)

    V = V + dt * f32(0.5) * (F_old + F)

    # NOTE(schsam): This will be wrong if F_norm ~< 1e-8.
    # TODO(schsam): We should check for forces below 1e-6. @ErrorChecking
    F_norm = jnp.sqrt(jnp.sum(F ** f32(2)) + f32(1e-6))
    V_norm = jnp.sqrt(jnp.sum(V ** f32(2)))

    P = jnp.array(jnp.dot(jnp.reshape(F, (-1)), jnp.reshape(V, (-1))))

    V = V + alpha * (F * V_norm / F_norm - V)

    # NOTE(schsam): Can we clean this up at all?
    n_pos = jnp.where(P >= 0, n_pos + 1, 0)
    dt_choice = jnp.array([dt * f_inc, dt_max])
    dt = jnp.where(P > 0,
                   jnp.where(n_pos > n_min,
                             jnp.min(dt_choice),
                             dt),
                   dt)
    dt = jnp.where(P < 0, dt * f_dec, dt)
    alpha = jnp.where(P > 0,
                      jnp.where(n_pos > n_min,
                                alpha * f_alpha,
                                alpha),
                      alpha)
    alpha = jnp.where(P < 0, alpha_start, alpha)
    V = (P < 0) * jnp.zeros_like(V) + (P >= 0) * V

    return FireDescentState(R, V, F, dt, alpha, n_pos)  # pytype: disable=wrong-arg-count
  return init_fn, apply_fn



# Copyright 2019 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Describes different physical quantities."""


from typing import TypeVar, Callable, Union, Tuple, Optional

from jax import grad, vmap, eval_shape
import jax.numpy as jnp
from jax import ops

from jax_md import space, dataclasses, partition, util

from functools import partial


# Types


Array = util.Array
f32 = util.f32
f64 = util.f64

DisplacementFn = space.DisplacementFn
MetricFn = space.MetricFn
Box = space.Box

EnergyFn = Callable[..., Array]
ForceFn = Callable[..., Array]

T = TypeVar('T')
InitFn = Callable[..., T]
ApplyFn = Callable[[T], T]
Simulator = Tuple[InitFn, ApplyFn]


# Functions


def force(energy_fn: EnergyFn) -> ForceFn:
  """Computes the force as the negative gradient of an energy."""
  return grad(lambda R, *args, **kwargs: -energy_fn(R, *args, **kwargs))


def clipped_force(energy_fn: EnergyFn, max_force: float) -> ForceFn:
  force_fn = force(energy_fn)
  def wrapped_force_fn(R, *args, **kwargs):
    force = force_fn(R, *args, **kwargs)
    force_norm = jnp.linalg.norm(force, axis=-1, keepdims=True)
    return jnp.where(force_norm > max_force,
                     force / force_norm * max_force,
                     force)

  return wrapped_force_fn

def canonicalize_force(energy_or_force_fn: Union[EnergyFn, ForceFn]
                       ) -> ForceFn:
  """Returns a force function given either an energy or force function."""
  _force_fn = None
  def force_fn(position, *args, **kwargs):
    nonlocal _force_fn
    if _force_fn is None:
      out_shape = eval_shape(energy_or_force_fn,
                             position, *args, **kwargs).shape
      if out_shape == ():
        _force_fn = force(energy_or_force_fn)
      elif out_shape == position.shape:
        _force_fn = energy_or_force_fn
      else:
        raise ValueError('Provided function should be compatible with either '
                         'an energy or a force. Found a function whose output '
                         f'has shape {out_shape}.')
    return _force_fn(position, *args, **kwargs)

  return force_fn


def volume(dimension: int, box: Box) -> float:
  if jnp.isscalar(box) or not box.ndim:
    return box ** dimension
  elif box.ndim == 1:
    return jnp.prod(box)
  elif box.ndim == 2:
    return jnp.linalg.det(box)
  raise ValueError(('Box must be either: a scalar, a vector, or a matrix. '
                    f'Found {box}.'))


def kinetic_energy(velocity: Array, mass: Array=1.0) -> float:
  """Computes the kinetic energy of a system with some velocities."""
  return 0.5 * util.high_precision_sum(mass * velocity ** 2)


def temperature(velocity: Array, mass: Array=1.0) -> float:
  """Computes the temperature of a system with some velocities."""
  N, dim = velocity.shape
  return util.high_precision_sum(mass * velocity ** 2) / (N * dim)


def pressure(energy_fn: EnergyFn, position: Array, box: Box,
             kinetic_energy: float=0.0, **kwargs) -> float:
  """Computes the internal pressure of a system.

  Args:
    energy_fn: A function that computes the energy of the system. This
      function must take as an argument `perturbation` which perturbs the
      box shape. Any energy function constructed using `smap` or in `energy.py`
      with a standard space will satisfy this property.
    position: An array of particle positions.
    box: A box specifying the shape of the simulation volume. Used to infer the
      volume of the unit cell.
    kinetic_energy: A float specifying the kinetic energy of the system.

  Returns:
    A float specifying the pressure of the system.
  """
  dim = position.shape[1]

  def U(eps):
    try:
      return energy_fn(position, box=box, perturbation=(1 + eps), **kwargs)
    except space.UnexpectedBoxException:
      return energy_fn(position, perturbation=(1 + eps), **kwargs)

  dUdV = grad(U)
  vol_0 = volume(dim, box)

  return 1 / (dim * vol_0) * (2 * kinetic_energy - dUdV(0.0))


def stress(energy_fn: EnergyFn, position: Array, box: Box,
           mass: Array=1.0, velocity: Optional[Array]=None, **kwargs
           ) -> Array:
  """Computes the internal stress of a system.

  Args:
    energy_fn: A function that computes the energy of the system. This
      function must take as an argument `perturbation` which perturbs the
      box shape. Any energy function constructed using `smap` or in `energy.py`
      with a standard space will satisfy this property.
    position: An array of particle positions.
    box: A box specifying the shape of the simulation volume. Used to infer the
      volume of the unit cell.
    mass: The mass of the particles; only used to compute the kinetic
      contribution if `velocity` is not `None`.
    velocity: An array of atomic velocities.

  Returns:
    A float specifying the pressure of the system.
  """
  dim = position.shape[1]

  zero = jnp.zeros((dim, dim), position.dtype)
  I = jnp.eye(dim, dtype=position.dtype)

  def U(eps):
    try:
      return energy_fn(position, box=box, perturbation=(I + eps), **kwargs)
    except space.UnexpectedBoxException:
      return energy_fn(position, perturbation=(I + eps), **kwargs)

  dUdV = grad(U)
  vol_0 = volume(dim, box)

  VxV = 0.0
  if velocity is not None:
    V = velocity
    VxV = util.high_precision_sum(mass * V[:, None, :] * V[:, :, None], axis=0)

  return 1 / vol_0 * (VxV - dUdV(zero))


def canonicalize_mass(mass: Union[float, Array]) -> Union[float, Array]:
  if isinstance(mass, float):
    return mass
  elif isinstance(mass, jnp.ndarray):
    if len(mass.shape) == 2 and mass.shape[1] == 1:
      return mass
    elif len(mass.shape) == 1:
      return jnp.reshape(mass, (mass.shape[0], 1))
    elif len(mass.shape) == 0:
      return mass
  elif isinstance(mass, f32) or isinstance(mass, f64):
    return mass
  msg = (
      'Expected mass to be either a floating point number or a one-dimensional'
      'ndarray. Found {}.'.format(mass)
      )
  raise ValueError(msg)


def cosine_angle_between_two_vectors(dR_12: Array, dR_13: Array) -> Array:
  dr_12 = space.distance(dR_12) + 1e-7
  dr_13 = space.distance(dR_13) + 1e-7
  cos_angle = jnp.dot(dR_12, dR_13) / dr_12 / dr_13
  return jnp.clip(cos_angle, -1.0, 1.0)


def cosine_angles(dR: Array) -> Array:
  """Returns cosine of angles for all atom triplets.

  Args:
    dR: Matrix of displacements; `ndarray(shape=[num_atoms, num_neighbors,
      spatial_dim])`.

  Returns:
    Tensor of cosine of angles;
    `ndarray(shape=[num_atoms, num_neighbors, num_neighbors])`.
  """

  angles_between_all_triplets = vmap(
      vmap(vmap(cosine_angle_between_two_vectors, (0, None)), (None, 0)), 0)
  return angles_between_all_triplets(dR, dR)


def is_integer(x: Array) -> bool:
  return x.dtype == jnp.int32 or x.dtype == jnp.int64


def pair_correlation(displacement_or_metric: Union[DisplacementFn, MetricFn],
                     radii: Array,
                     sigma: float,
                     species: Array = None,
                     eps: float = 1e-7):
  """Computes the pair correlation function at a mesh of distances.

  The pair correlation function measures the number of particles at a given
  distance from a central particle. The pair correlation function is defined by

  .. math::
    g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>
  
  We make the approximation,

  .. math::
    \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

  Args:
    displacement_or_metric: 
      A function that computes the displacement or distance between two points.
    radii: An array of radii at which we would like to compute :math:`g(r)`.
    sigima: A float specifying the width of the approximating Gaussian.
    species: An optional array specifying the species of each particle. If
      species is None then we compute a single :math:`g(r)` for all particles,
      otherwise we compute one :math:`g(r)` for each species.
    eps: A small additive constant used to ensure stability if the radius is
      zero.

  Returns:
    A function `g_fn` that computes the pair correlation function for a
    collection of particles.
  """
  d = space.canonicalize_displacement_or_metric(displacement_or_metric)
  d = space.map_product(d)

  inv_rad = 1 / (radii + eps)

  def pairwise(dr, dim):
    return jnp.exp(-f32(0.5) * (dr - radii)**2 / sigma**2) * inv_rad**(dim - 1)
  pairwise = vmap(vmap(pairwise, (0, None)), (0, None))

  if species is None:
    def g_fn(R):
      dim = R.shape[-1]
      mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
      return jnp.sum(mask[:, :, jnp.newaxis] *
                     pairwise(d(R, R), dim), axis=(1,))
  else:
    if not (isinstance(species, jnp.ndarray) and is_integer(species)):
      raise TypeError('Malformed species; expecting array of integers.')
    species_types = jnp.unique(species)
    def g_fn(R):
      dim = R.shape[-1]
      g_R = []
      mask = 1 - jnp.eye(R.shape[0], dtype=R.dtype)
      for s in species_types:
        Rs = R[species == s]
        mask_s = mask[:, species == s, jnp.newaxis]
        g_R += [jnp.sum(mask_s * pairwise(d(Rs, R), dim), axis=(1,))]
      return g_R
  return g_fn


def pair_correlation_neighbor_list(
    displacement_or_metric: Union[DisplacementFn, MetricFn],
    box_size: Box,
    radii: Array,
    sigma: float,
    species: Array = None,
    dr_threshold: float = 0.5,
    eps: float = 1e-7,
    fractional_coordinates: bool=False,
    format: partition.NeighborListFormat=partition.Dense):
  """Computes the pair correlation function at a mesh of distances.

  The pair correlation function measures the number of particles at a given
  distance from a central particle. The pair correlation function is defined by 
  
  .. math::
    g(r) = <\sum_{i \\neq j}\delta(r - |r_i - r_j|)>
  
  We make the approximation,

  .. math::
    \delta(r) \\approx {1 \over \sqrt{2\pi\sigma^2}e^{-r / (2\sigma^2)}}

  This function uses neighbor lists to speed up the calculation.

  Args:
    displacement_or_metric: A function that computes the displacement or
      distance between two points.
    box_size: The size of the box containing the particles.
    radii: An array of radii at which we would like to compute :math:`g(r)`.
    sigima: A float specifying the width of the approximating Gaussian.
    species: An optional array specifying the species of each particle. If
      species is None then we compute a single :math:`g(r)` for all particles,
      otherwise we compute one :math:`g(r)` for each species.
    dr_threshold: A float specifying the halo size of the neighbor list.
    eps: A small additive constant used to ensure stability if the radius is
      zero.
    fractional_coordinates: Bool determining whether positions are stored in
      the unit cube or not.
    format: The format of the neighbor lists. Must be `Dense` or `Sparse`.

  Returns:
    A pair of functions: `neighbor_fn` that constructs a neighbor list (see
    `neighbor_list` in `partition.py` for details). `g_fn` that computes the
    pair correlation function for a collection of particles given their
    position and a neighbor list.
  """
  metric = space.canonicalize_displacement_or_metric(displacement_or_metric)
  inv_rad = 1 / (radii + eps)
  def pairwise(dr, dim):
    return jnp.exp(-f32(0.5) * (dr - radii)**2 / sigma**2) * inv_rad**(dim - 1)

  neighbor_fn = partition.neighbor_list(displacement_or_metric,
                                        box_size,
                                        jnp.max(radii) + sigma,
                                        dr_threshold,
                                        format=format)

  if species is None:
    def g_fn(R, neighbor):
      N, dim = R.shape
      mask = partition.neighbor_list_mask(neighbor)
      if neighbor.format is partition.Dense:
        R_neigh = R[neighbor.idx]
        d = space.map_neighbor(metric)
        _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
        return jnp.sum(mask[:, :, None] *
                       _pairwise(d(R, R_neigh), dim), axis=(1,))
      elif neighbor.format is partition.Sparse:
        dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
        _pairwise = vmap(pairwise, (0, None))
        return ops.segment_sum(mask[:, None] * _pairwise(dr, dim),
                               neighbor.idx[0],
                               N)
      else:
        raise NotImplementedError('Pair correlation function does not support '
                                  'OrderedSparse neighbor lists.')

  else:
    if not (isinstance(species, jnp.ndarray) and is_integer(species)):
      raise TypeError('Malformed species; expecting array of integers.')
    species_types = jnp.unique(species)
    def g_fn(R, neighbor):
      N, dim = R.shape
      g_R = []
      mask = partition.neighbor_list_mask(neighbor)
      if neighbor.format is partition.Dense:
        neighbor_species = species[neighbor.idx]
        R_neigh = R[neighbor.idx]
        d = space.map_neighbor(metric)
        _pairwise = vmap(vmap(pairwise, (0, None)), (0, None))
        for s in species_types:
          mask_s = mask * (neighbor_species == s)
          g_R += [jnp.sum(mask_s[:, :, jnp.newaxis] *
                          _pairwise(d(R, R_neigh), dim), axis=(1,))]
      elif neighbor.format is partition.Sparse:
        neighbor_species = species[neighbor.idx[1]]
        dr = space.map_bond(metric)(R[neighbor.idx[0]], R[neighbor.idx[1]])
        _pairwise = vmap(pairwise, (0, None))
        for s in species_types:
          mask_s = mask * (neighbor_species == s)
          g_R += [ops.segment_sum(mask_s[:, None] *
                                  _pairwise(dr, dim), neighbor.idx[0], N)]
      else:
        raise NotImplementedError('Pair correlation function does not support '
                                  'OrderedSparse neighbor lists.')

      return g_R
  return neighbor_fn, g_fn


def box_size_at_number_density(particle_count: int,
                               number_density: float,
                               spatial_dimension: int) -> float:
  return jnp.power(particle_count / number_density, 1 / spatial_dimension)


def bulk_modulus(elastic_tensor: Array) -> float:
  return jnp.einsum('iijj->', elastic_tensor) / elastic_tensor.shape[0] ** 2


@dataclasses.dataclass
class PHopState:
    position_buffer: jnp.ndarray
    phop: jnp.ndarray

InitFn = Callable[[Array], PHopState]
ApplyFn = Callable[[PHopState, Array], PHopState]
PHopCalculator = Tuple[InitFn, ApplyFn]

def phop(displacement: DisplacementFn, window_size: int) -> PHopCalculator:
  """Computes the phop indicator of rearrangements.

  phop is an indicator function that is effective at detecting when particles
  in a quiescent system have experienced a rearrangement. Qualitatively, phop
  measures when the average position of a particle has changed significantly.

  Formally, given a window of size :math:`\Delta t` we two averages before and after
  the current time,

  .. math::
    E_A[f] = E_{t\in[t - \Delta t / 2, t]}[f(t)]
    E_B[f] = E_{t\in[t, t + \Delta t / 2]}[f(t)].

  In terms of these expectations, phop is given by,

  .. math::
    phop = \sqrt{E_A[(R_i(t) - E_B[R_i(t)])^2]E_B[(R_i(t) - E_A[R_i(t)])^2]}

  phop was first introduced in Candelier et al. [#candelier]_

  Args:
    displacement: A function that computes displacements between pairs of
      particles. See `spaces.py` for details.
    window_size: An integer specifying the number of positions that constitute
      the window.

  Returns:
    A pair of functions, `(init_fn, update_fn)` that initialize the state of a
    phop measurement and update the state of a phop measurement to include new
    positions.

  .. rubric:: References
  .. [#candelier] R. Candelier et al.
    "Spatiotemporal Hierarchy of Relaxation Events, Dynamical Heterogeneities,
    and Structural Reorganization in a Supercooled Liquid"
    Physical Review Letters 105, 135702 (2010).
  """

  half_window_size = window_size // 2
  displacement = space.map_bond(displacement)

  def init_fn(position: Array) -> PHopState:
    position_buffer = jnp.tile(position, (window_size, 1, 1))
    assert position_buffer.shape == ((window_size,) + position.shape)
    return PHopState(position_buffer, jnp.zeros((position.shape[0],)))  # pytype: disable=wrong-arg-count

  def update_fn(state: PHopState, position: Array) -> PHopState:
    # Compute phop.
    a_pos = state.position_buffer[:half_window_size]
    a_mean = jnp.mean(a_pos, axis=0)
    b_pos = state.position_buffer[half_window_size:]
    b_mean = jnp.mean(b_pos, axis=0)

    phop = jnp.sqrt(jnp.mean((a_pos - b_mean) ** 2 * (b_pos - a_mean) ** 2,
                             axis=(0, 2)))

    # Unwrap position.
    buff = state.position_buffer
    position = displacement(position, buff[-1]) + buff[-1]

    # Add position to the list.
    buff = jnp.concatenate((buff, position[jnp.newaxis, :, :]))[1:]

    return PHopState(buff, phop)  # pytype: disable=wrong-arg-count

  return init_fn, update_fn
